import logging
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

import librosa
import matplotlib.pyplot as plt
import numpy as np
import soundfile as sf
import torch
from numpy.typing import NDArray
from torch import Tensor

from audiozen.constant import EPSILON


logger = logging.getLogger(__name__)


def find_files(path_or_path_list: Union[list, str, Path], offset: int = 0, limit: Union[int, None] = None):
    """Find wav files from a directory, or a list of files, or a txt file, or the combination of them.

    Args:
        path_or_path_list: path to wav file, str, pathlib.Path, or a list of them.
        offset: offset of samples to load.
        limit: limit of samples to load.

    Returns:
        A list of wav file paths.

    Examples:
        >>> # Load 10 files from a directory
        >>> wav_paths = file_loader(path="~/dataset", limit=10, offset=0)
        >>> # Load files from a directory, a txt file, and a wav file
        >>> wav_paths = file_loader(path=["~/dataset", "~/scp.txt", "~/test.wav"])
    """
    if not isinstance(path_or_path_list, list):
        path_or_path_list = [path_or_path_list]

    output_paths = []
    for path in path_or_path_list:
        path = Path(path).resolve()

        if path.is_dir():
            wav_paths = librosa.util.find_files(path, ext="wav")
            output_paths += wav_paths

        if path.is_file():
            if path.suffix == ".wav":
                output_paths.append(path.as_posix())
            else:
                for line in open(path, "r"):
                    line = line.rstrip("\n")
                    line = Path(line).resolve()
                    output_paths.append(line.as_posix())

    if offset > 0:
        output_paths = output_paths[offset:]

    if limit:
        output_paths = output_paths[:limit]

    return output_paths


def is_audio(y_s: Sequence[NDArray] | NDArray):
    if not isinstance(y_s, Sequence):
        y_s = [y_s]

    for y in y_s:
        assert y.ndim in (1, 2), "Only support signals with the shape of [C, T] or [T]."


def compute_rms(y: NDArray) -> float:
    """Compute the Root Mean Square (RMS) of the given signal."""
    return np.sqrt(np.mean(y**2))


def loudness_max_norm(
    y: NDArray, *, scalar: Optional[float] = None, ref_mic: int = 0, eps: float = EPSILON
) -> tuple[NDArray, float]:
    """Maximum loudness normalization to signals."""
    if scalar is None:
        if y.ndim == 1:
            scalar = np.max(np.abs(y)) + eps
        else:
            scalar = np.max(np.abs(y[ref_mic, :])) + eps

    assert scalar is not None
    return y / scalar, scalar


def loudness_rms_norm(
    y: NDArray, scalar: Union[float, None] = None, lvl: float = -25, ref_mic: int = 0, eps: float = EPSILON
) -> tuple[NDArray, float]:
    """Loudness normalize a signal based on the Root Mean Square (RMS).

    Normalize the RMS of signals to a given RMS based on Decibels Relative to Full Scale (dBFS).

    Args:
        y: [C, T] or [T,].
        scalar: scalar to normalize the RMS, default to None.
        target_rms: target RMS in dBFS.
        ref_mic: reference mic for multi-channel signals.

    Returns:
        Loudness normalized signal and scalar.

    Note:
        A small amount of signal samples would be clipped after normalization, but it does not matter.
    """
    if not scalar:
        current_level = compute_rms(y) if y.ndim == 1 else compute_rms(y[ref_mic, :])
        scalar = 10 ** (lvl / 20) / (current_level + eps)

    return y * scalar, scalar


def active_rms(clean, noise, sr=16000, energy_threshold=-50, eps=EPSILON):
    """Compute the active RMS of clean and noise signals based on the energy threshold (dB)."""
    window_size = 100  # ms
    window_samples = int(sr * window_size / 1000)
    sample_start = 0

    noise_active_segments = np.array([])
    clean_active_segments = np.array([])

    while sample_start < len(noise):
        sample_end = min(sample_start + window_samples, len(noise))
        noise_win = noise[sample_start:sample_end]
        clean_win = clean[sample_start:sample_end]
        noise_seg_rms = compute_rms(noise_win)

        if noise_seg_rms > energy_threshold:
            noise_active_segments = np.append(noise_active_segments, noise_win)
            clean_active_segments = np.append(clean_active_segments, clean_win)

        sample_start += window_samples

    if len(noise_active_segments) != 0:
        noise_rms = compute_rms(noise_active_segments)
    else:
        noise_rms = eps

    if len(clean_active_segments) != 0:
        clean_rms = compute_rms(clean_active_segments)
    else:
        clean_rms = eps

    return clean_rms, noise_rms


def normalize_segmental_rms(audio, rms, target_lvl=-25, eps=EPSILON):
    """Normalize the RMS of a segment to a target level.

    Args:
        audio: audio segment.
        rms: RMS of the audio segment.
        target_lvl: target level in dBFS.
        eps: a small value to avoid dividing by zero.

    Returns:
        Normalized audio segment.
    """
    scalar = 10 ** (target_lvl / 20) / (rms + eps)
    return audio * scalar


def sxr2gain(meaningful: NDArray, meaningless: NDArray, desired_ratio: float, eps: float = EPSILON) -> float:
    """Generally calculate the gains of interference to fulfill a desired SXR (SNR or SIR) ratio.

    Args:
        meaningful: meaningful input, like target clean.
        meaningless: meaningless or unwanted input, like background noise.
        desired_ratio: SNR or SIR ratio.

    Returns:
        Gain, which can be used to adjust the RMS of the meaningless signals to satisfy the given ratio.
    """
    meaningful_rms = compute_rms(meaningful)
    meaningless_rms = compute_rms(meaningless)
    scalar = meaningful_rms / (10 ** (desired_ratio / 20)) / (meaningless_rms + eps)

    return scalar


def load_wav(wav_path: Union[str, Path], *, sr: int = 16000) -> NDArray:
    """Load a wav file.

    Args:
        file: file path.
        sr: sample rate. Defaults to 16000.

    Returns:
        Waveform with shape of [C, T] or [T].
    """
    wav_path = Path(wav_path).resolve()
    y, _ = librosa.load(wav_path, sr=sr, mono=False)
    return y


def save_wav(data: NDArray | Tensor, fpath: Union[Path, list], sr: int):
    if data.ndim != 1:
        data = data.reshape(-1)

    if isinstance(data, Tensor):
        data = data.detach().cpu().numpy()

    sf.write(fpath, data, sr)


def mag_phase(complex_valued_tensor: Tensor) -> tuple[Tensor, Tensor]:
    """Get magnitude and phase of a complex-valued tensor.

    Args:
        complex_valued_tensor: complex-valued tensor.

    Returns:
        magnitude and phase spectrogram.
    """
    mag, phase = torch.abs(complex_valued_tensor), torch.angle(complex_valued_tensor)
    return mag, phase


def mag_phase_to_real_imag(mag: Tensor, phase: Tensor) -> tuple[Tensor, Tensor]:
    """Convert magnitude and phase spectrogram to real and imag spectrogram.

    Args:
        mag: magnitude spectrogram.
        phase: phase spectrogram.

    Returns:
        real and imag spectrogram.
    """
    return mag * torch.cos(phase), mag * torch.sin(phase)


def mag_phase_to_complex(mag: Tensor, phase: Tensor) -> Tensor:
    """Convert magnitude and phase spectrogram to complex-valued tensor.

    Args:
        mag: magnitude spectrogram.
        phase: phase spectrogram.

    Returns:
        complex-valued tensor.
    """
    return mag * torch.exp(1j * phase)


def real_imag(complex_valued_tensor: Tensor) -> tuple[Tensor, Tensor]:
    """Get real and imag part of a complex-valued tensor.

    Args:
        complex_valued_tensor: complex-valued tensor.

    Returns:
        real and imag spectrogram.
    """
    return complex_valued_tensor.real, complex_valued_tensor.imag


def complex_view_as_real(complex_valued_tensor: Tensor) -> Tensor:
    """View a complex-valued tensor as a real-valued tensor.

    Args:
        complex_valued_tensor: complex-valued tensor.

    Returns:
        A real-valued tensor with shape of [B, 2, F, T].
    """
    return torch.stack((complex_valued_tensor.real, complex_valued_tensor.imag), dim=1)


def real_imag_to_mag_phase(real: Tensor, imag: Tensor) -> tuple[Tensor, Tensor]:
    """Convert real and imag spectrogram to magnitude and phase spectrogram.

    Args:
        real: real spectrogram.
        imag: imag spectrogram.

    Returns:
        magnitude and phase spectrogram.
    """
    return mag_phase(torch.complex(real=real, imag=imag))


def stft_v2(
    y: Tensor,
    n_fft: int,
    hop_length: int,
    win_length: int,
    window: Optional[str] = "hann",
    **kwargs,
) -> Tensor:
    """Wrapper of the official ``torch.stft`` for single-channel and multichannel signals.

    Args:
        y (`torch.Tensor` of shape `(batch_size, num_channels, num_samples) or `(batch_size, num_samples)`):
            single-/multichannel signals.
        n_fft: num of FFT.
        hop_length: hop length.
        win_length: hanning window size.
        kwargs: other arguments for ``torch.stft``.

    Returns:
        If the input is single-channel, return the spectrogram with shape of [B, F, T], otherwise [B, C, F, T].
    """
    # Support [B, T] and [B, C, T]
    if y.ndim not in [2, 3]:
        raise ValueError(f"Only support single-/multi-channel signals. Received {y.ndim=}.")

    is_multi_channel = y.ndim == 3

    batch_size, *_, num_samples = y.shape

    # Compatible with multi-channel signals
    if is_multi_channel:
        y = y.reshape(-1, num_samples)

    if window is not None:
        window_func = getattr(torch, f"{window}_window")
        window_tensor = window_func(win_length, device=y.device)
    else:
        window_tensor = torch.ones(n_fft, device=y.device)

    complex_stft = torch.stft(
        input=y,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window_tensor,
        return_complex=True,
        **kwargs,
    )

    # Reshape back to original if the input is multi-channel
    if is_multi_channel:
        complex_stft = complex_stft.reshape(batch_size, -1, *complex_stft.shape[-2:])

    return complex_stft


def stft(
    y,
    n_fft,
    hop_length,
    win_length,
    window="hann",
    output_type: Literal["mag_phase", "real_imag", "complex"] | None = None,
    **kwargs,
) -> tuple[Tensor, ...] | Tensor:
    """Wrapper of the official ``torch.stft`` for single-channel and multichannel signals.

    Args:
        y (`torch.Tensor` of shape `(batch_size, num_channels, num_samples) or `(batch_size, num_samples)`):
            single-/multichannel signals.
        n_fft: num of FFT.
        hop_length: hop length.
        win_length: hanning window size.
        output_type: "mag_phase", "real_imag", "complex", or None. Defaults to None.
        kwargs: other arguments for ``torch.stft``.

    Returns:
        If the input is single-channel, return the spectrogram with shape of [B, F, T], otherwise [B, C, F, T].
        If output_type is "mag_phase", return a list of magnitude and phase spectrogram.
        If output_type is "real_imag", return a list of real and imag spectrogram.
        If output_type is None, return a list of magnitude, phase, real, and imag spectrogram.
    """
    warnings.warn("This function will be deprecated in the future.", DeprecationWarning, stacklevel=2)

    # Support [B, T] and [B, C, T]
    if y.ndim not in [2, 3]:
        raise ValueError(f"Only support single-/multi-channel signals. Received {y.ndim=}.")

    is_multi_channel = y.ndim == 3

    batch_size, *_, num_samples = y.shape

    # Compatible with multi-channel signals
    if is_multi_channel:
        y = y.reshape(-1, num_samples)

    if window is not None:
        window_func = getattr(torch, f"{window}_window")
        window = window_func(win_length, device=y.device)
    else:
        window = torch.ones(n_fft, device=y.device)

    complex_stft = torch.stft(
        input=y,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window,
        return_complex=True,
        **kwargs,
    )

    # Reshape back to original if the input is multi-channel
    if is_multi_channel:
        complex_stft = complex_stft.reshape(batch_size, -1, *complex_stft.shape[-2:])

    if output_type == "mag_phase":
        return mag_phase(complex_stft)
    elif output_type == "real_imag":
        return real_imag(complex_stft)
    elif output_type == "complex":
        return complex_stft
    else:
        mag, phase = mag_phase(complex_stft)
        return mag, phase, complex_stft.real, complex_stft.imag


def istft_v2(
    complex_valued_features: Tensor,
    n_fft: int,
    hop_length: int,
    win_length: int,
    window: Optional[str] = "hann",
    length: Optional[int] = None,
) -> Tensor:
    """Wrapper of the official ``torch.istft`` for single-channel and multichannel signals.

    Args:
        complex_valued_features: complex-valued tensor.
        n_fft: num of FFT.
        hop_length: hop length.
        win_length: hanning window size.
        window: window function.
        length: length of the output signal.

    Returns:
        Time-domain signal with shape of [B, T] or [B, C, T].
    """
    # Support [B, F, T] and [B, C, F, T]
    if complex_valued_features.ndim not in [3, 4]:
        raise ValueError(f"Only support single-/multi-channel signals. Received {complex_valued_features.ndim=}.")

    is_multi_channel = complex_valued_features.ndim == 4

    batch_size, *_, num_frames = complex_valued_features.shape

    # Compatible with multi-channel signals
    if is_multi_channel:
        complex_valued_features = complex_valued_features.reshape(-1, *complex_valued_features.shape[-2:])

    if window is not None:
        window_func = getattr(torch, f"{window}_window")
        window_initialized = window_func(win_length, device=complex_valued_features.device)
    else:
        window_initialized = torch.ones(win_length, device=complex_valued_features.device)

    time_domain = torch.istft(
        input=complex_valued_features,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window_initialized,
        length=length,
    )

    # Reshape back to original if the input is multi-channel
    if is_multi_channel:
        time_domain = time_domain.reshape(batch_size, -1, time_domain.shape[-1])

    return time_domain


def istft(
    feature: Union[Tensor, Tuple[Tensor, Tensor], List[Tensor]],
    n_fft: int,
    hop_length: int,
    win_length: int,
    window: Optional[str] = "hann",
    length: Optional[int] = None,
    input_type: Literal["mag_phase", "real_imag", "complex", "complex_view_as_real"] = "mag_phase",
) -> Tensor:
    warnings.warn("This function will be deprecated in the future.", DeprecationWarning, stacklevel=2)

    if input_type == "real_imag":
        if isinstance(feature, (tuple, list)) and len(feature) == 2:
            real, imag = feature
            complex_valued_features = torch.complex(real=real, imag=imag)
        else:
            raise ValueError("For 'real_imag', provide a tuple or list with two elements: (real, imag).")

    elif input_type == "complex":
        if isinstance(feature, Tensor) and torch.is_complex(feature):
            complex_valued_features = feature
        else:
            raise ValueError("For 'complex', provide a complex-valued tensor.")

    elif input_type == "complex_view_as_real":
        assert isinstance(feature, Tensor), "For 'complex_view_as_real', provide a complex-valued tensor."
        if feature.ndim == 4 and feature.shape[1] == 2:
            real, imag = feature[:, 0, ...], feature[:, 1, ...]
            complex_valued_features = torch.complex(real=real, imag=imag)
        else:
            raise ValueError("For 'complex_view_as_real', provide a 4D tensor with shape [B, 2, F, T].")

    elif input_type == "mag_phase":
        if isinstance(feature, (tuple, list)) and len(feature) == 2:
            mag, phase = feature
            complex_valued_features = torch.polar(mag, phase)
        else:
            raise ValueError("For 'mag_phase', provide a tuple or list with two elements: (magnitude, phase).")

    else:
        raise ValueError(f"Unsupported input_type: {input_type}")

    if window is not None:
        window_func = getattr(torch, f"{window}_window", None)
        if window_func is None:
            raise ValueError(f"Unsupported window type: {window}")
        window_initialized = window_func(win_length, device=complex_valued_features.device)
    else:
        window_initialized = torch.ones(win_length, device=complex_valued_features.device)

    return torch.istft(
        input=complex_valued_features,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window_initialized,
        length=length,
    )


def norm_amplitude(y: NDArray, scalar: Optional[float] = None, eps: float = EPSILON):
    if not scalar:
        scalar = np.max(np.abs(y)) + eps

    return y / scalar, scalar


def is_clipped(y: NDArray, clipping_threshold: float = 0.999) -> bool:
    """Check if the input signal is clipped."""
    return (np.abs(y) > clipping_threshold).any()  # type: ignore


def tune_dB_FS(y, target_dB_FS=-26, eps=EPSILON):
    """Tune the RMS of the input signal to a target level.

    Args:
        y: Audio signal with any shape.
        target_dB_FS: Target dB_FS. Defaults to -25.
        eps: A small value to avoid dividing by zero. Defaults to EPSILON.

    Returns:
        Scaled audio signal with the same shape as the input.
    """
    if isinstance(y, torch.Tensor):
        rms = torch.sqrt(torch.mean(y**2))
        scalar = 10 ** (target_dB_FS / 20) / (rms + eps)
        y *= scalar
        return y, rms, scalar
    else:
        rms = np.sqrt(np.mean(y**2))
        scalar = 10 ** (target_dB_FS / 20) / (rms + eps)
        y *= scalar
        return y, rms, scalar


def activity_detector(audio, fs=16000, activity_threshold=0.13, target_level=-25, eps=EPSILON):
    """Return the percentage of the time the audio signal is above an energy threshold.

    Args:
        audio:
        fs:
        activity_threshold:
        target_level:
        eps:

    Returns:
    """
    audio, _ = loudness_rms_norm(audio, target_level)
    window_size = 50  # ms
    window_samples = int(fs * window_size / 1000)
    sample_start = 0
    cnt = 0
    prev_energy_prob = 0
    active_frames = 0

    a = -1
    b = 0.2
    alpha_rel = 0.05
    alpha_att = 0.8

    while sample_start < len(audio):
        sample_end = min(sample_start + window_samples, len(audio))
        audio_win = audio[sample_start:sample_end]
        frame_rms = 20 * np.log10(sum(audio_win**2) + eps)
        frame_energy_prob = 1.0 / (1 + np.exp(-(a + b * frame_rms)))

        if frame_energy_prob > prev_energy_prob:
            smoothed_energy_prob = frame_energy_prob * alpha_att + prev_energy_prob * (1 - alpha_att)
        else:
            smoothed_energy_prob = frame_energy_prob * alpha_rel + prev_energy_prob * (1 - alpha_rel)

        if smoothed_energy_prob > activity_threshold:
            active_frames += 1
        prev_energy_prob = frame_energy_prob
        sample_start += window_samples
        cnt += 1

    perc_active = active_frames / cnt
    return perc_active


def build_complex_ideal_ratio_mask(noisy_real, noisy_imag, clean_real, clean_imag) -> torch.Tensor:
    """Build the complex ratio mask.

    Args:
        noisy: [B, F, T], noisy complex-valued stft coefficients
        clean: [B, F, T], clean complex-valued stft coefficients

    References:
        https://ieeexplore.ieee.org/document/7364200

    Returns:
        [B, F, T, 2]
    """
    denominator = torch.square(noisy_real) + torch.square(noisy_imag) + EPSILON

    mask_real = (noisy_real * clean_real + noisy_imag * clean_imag) / denominator
    mask_imag = (noisy_real * clean_imag - noisy_imag * clean_real) / denominator

    complex_ratio_mask = torch.stack((mask_real, mask_imag), dim=-1)

    return compress_cIRM(complex_ratio_mask, K=10, C=0.1)


def compress_cIRM(mask, K=10, C=0.1):
    """Compress the value of cIRM from (-inf, +inf) to [-K ~ K].

    References:
        https://ieeexplore.ieee.org/document/7364200
    """
    if torch.is_tensor(mask):
        mask = -100 * (mask <= -100) + mask * (mask > -100)
        mask = K * (1 - torch.exp(-C * mask)) / (1 + torch.exp(-C * mask))
    else:
        mask = -100 * (mask <= -100) + mask * (mask > -100)
        mask = K * (1 - np.exp(-C * mask)) / (1 + np.exp(-C * mask))
    return mask


def decompress_cIRM(mask, K=10, limit=9.9):
    """Decompress cIRM from [-K ~ K] to [-inf, +inf].

    Args:
        mask: cIRM mask
        K: default 10
        limit: default 0.1

    References:
        https://ieeexplore.ieee.org/document/7364200
    """
    mask = limit * (mask >= limit) - limit * (mask <= -limit) + mask * (torch.abs(mask) < limit)
    mask = -K * torch.log((K - mask) / (K + mask))
    return mask


def complex_mul(noisy_r, noisy_i, mask_r, mask_i):
    r = noisy_r * mask_r - noisy_i * mask_i
    i = noisy_r * mask_i + noisy_i * mask_r
    return r, i


class Mask:
    def __init__(self) -> None:
        pass

    @staticmethod
    def plot_mask(mask, title="Mask", xlim=None):
        mask = mask.numpy()
        figure, axis = plt.subplots(1, 1)
        img = axis.imshow(mask, cmap="viridis", origin="lower", aspect="auto")
        figure.suptitle(title)
        plt.colorbar(img, ax=axis)
        plt.show()


class IRM(Mask):
    def __init__(self) -> None:
        super(IRM, self).__init__()

    @staticmethod
    def generate_mask(clean, noise, ref_channel=0):
        """Generate an ideal ratio mask.

        Args:
            clean: Complex STFT of clean audio with the shape of [B, C, F, T].
            noise: Complex STFT of noise audio with the same shape of [B, 1, F, T].
            ref_channel: The reference channel to compute the mask if the STFTs are multi-channel.

        Returns:
            Speech mask and noise mask with the shape of [B, 1, F, T].
        """
        mag_clean = clean.abs() ** 2
        mag_noise = noise.abs() ** 2
        irm_speech = mag_clean / (mag_clean + mag_noise)
        irm_noise = mag_noise / (mag_clean + mag_noise)

        if irm_speech.ndim == 4:
            irm_speech = irm_speech[ref_channel, :][:, None, ...]
            irm_noise = irm_noise[ref_channel, :][:, None, ...]

        return irm_speech, irm_noise


def drop_band(input, num_groups=2):
    """Reduce computational complexity of the sub-band part in the FullSubNet model.

    Shapes:
        input: [B, C, F, T]
        return: [B, C, F // num_groups, T]
    """
    batch_size, _, num_freqs, _ = input.shape
    assert batch_size > num_groups, (
        f"Batch size = {batch_size}, num_groups = {num_groups}. The batch size should larger than the num_groups."
    )

    if num_groups <= 1:
        # No demand for grouping
        return input

    # Each sample must has the same number of the frequencies for parallel training.
    # Therefore, we need to drop those remaining frequencies in the high frequency part.
    if num_freqs % num_groups != 0:
        input = input[..., : (num_freqs - (num_freqs % num_groups)), :]
        num_freqs = input.shape[2]

    output = []
    for group_idx in range(num_groups):
        samples_indices = torch.arange(group_idx, batch_size, num_groups, device=input.device)
        freqs_indices = torch.arange(group_idx, num_freqs, num_groups, device=input.device)

        selected_samples = torch.index_select(input, dim=0, index=samples_indices)
        selected = torch.index_select(selected_samples, dim=2, index=freqs_indices)  # [B, C, F // num_groups, T]

        output.append(selected)

    return torch.cat(output, dim=0)


def hanning_hugh(win_length):
    win = np.hanning(win_length)

    w_half = win[: win_length // 2]

    w_half = (w_half + (1 - np.flipud(w_half))) // 2

    w = np.zeros(win_length)
    w[: win_length // 2] = w_half
    w[-win_length // 2 :] = np.flipud(w_half)

    return w


def asymmetric_analysis_window(k, m, d=0):
    rising_sqrt_hann = np.sqrt(np.hanning(2 * (k - m - d) + 1)[: 2 * (k - m - d)])
    falling_sqrt_hann = np.sqrt(np.hanning(2 * m + 1)[: 2 * m])

    win = np.zeros(k)
    win[:d] = 0
    win[d : k - m] = rising_sqrt_hann[: k - m - d]
    win[k - m :] = falling_sqrt_hann[-m:]

    return win


def asymmetric_synthesis_window(k, m, d=0):
    rising_sqrt_hann = np.sqrt(np.hanning(2 * (k - m - d) + 1)[: 2 * (k - m - d)])
    rasing_norm_hann = np.hanning(2 * m + 1)[:m] / rising_sqrt_hann[k - 2 * m - d : k - m - d]
    falling_sqrt_hann = np.sqrt(np.hanning(2 * m + 1)[: 2 * m])

    win = np.zeros(k)
    win[: -2 * m] = 0
    win[-2 * m : -m] = rasing_norm_hann
    win[-m:] = falling_sqrt_hann[-m:]

    return win


def asymmetric_stft(y: Tensor, analysis_win: Union[Tensor, NDArray], win_length: int, hop_length: int):
    """Asymmetric short-time Fourier transform using PyTorch.

    Args:
        y: The input signal with a shape of `(batch_size, num_channels, num_samples)` or `(batch_size, num_samples)`.
        analysis_win: The analysis window function to apply to each frame.
        hop_length: The number of samples between successive frames.
        win_length: The length of the window function.

    Note:
        `num_windows x win_len` frames always longer than the original signal to make sure that the first and last
        samples are included in the output.

    Returns:
        Complex-valued matrix of short-term Fourier transform coefficients with a shape of (batch_size, num_freqs,
        num_frames).
    """
    # Ensure the input is 2D or 3D
    if y.ndim != 2 and y.ndim != 3:
        raise ValueError(f"Input signal must have shape [B, T] or [B, C, T], but received {y.shape=}.")

    # Ensure the length of the window function is equal to the window length
    if len(analysis_win) != win_length:
        raise ValueError(
            f"Length of the analysis window must be equal to the window length, but received "
            f"{len(analysis_win)=} and {win_length=}."
        )

    is_multi_channel = y.ndim == 3
    device = y.device

    origin_batch_size = None
    num_channels = None
    if is_multi_channel:
        origin_batch_size, num_channels, num_samples = y.shape
        y = y.reshape(-1, num_samples)

    if isinstance(analysis_win, np.ndarray):
        analysis_win = torch.tensor(analysis_win, dtype=torch.float32, device=device)

    # Pad the input signal to ensure that the first and last samples are included in the output although there are some
    # redundant computations
    y = torch.nn.functional.pad(y, (win_length, win_length), mode="constant")
    batch_size, num_samples = y.shape

    # Calculate the number of windows
    num_windows = ((num_samples - win_length) // hop_length) + 1

    # Create a complex-valued tensor to store the STFT
    stft_matrix = torch.empty((batch_size, num_windows, win_length // 2 + 1), dtype=torch.cfloat, device=device)

    # Create indices for sliding windows. This results in a shape of (win_len, num_windows) for all batches
    indices = torch.arange(win_length, device=device).unsqueeze(0) + torch.arange(
        0, num_windows * hop_length - 1, hop_length, device=device
    ).unsqueeze(1)

    # Gather the windows from each signal in the batch
    # Use `unsqueeze(1)` to replicate indices across the batch dimension
    windows = y[..., indices]  # Shape: (batch_size, win_len, num_windows)

    # Apply the window function to each of the gathered segments
    windowed_segments = analysis_win.unsqueeze(0) * windows  # Shape: (batch_size, win_len, num_windows)

    # Perform the FFT on all windowed segments in batch
    stft_matrix = torch.fft.rfft(
        windowed_segments.view(-1, win_length)
    )  # Reshape to (batch_size*num_windows, win_len)

    # Reshape the output to (batch_size, num_windows, win_len // 2 + 1)
    stft_matrix = stft_matrix.view(batch_size, num_windows, -1)
    stft_matrix = stft_matrix.permute(0, 2, 1).contiguous()

    if is_multi_channel:
        stft_matrix = stft_matrix.reshape(origin_batch_size, num_channels, -1, num_windows)

    return stft_matrix


def asymmetric_istft(stft_matrix, analysis_win, synthesis_win, win_length, hop_length, length=None):
    """Inverse STFT using PyTorch.

    Note:
        The input STFT matrix must have a shape of (num_freqs, num_frames).
        No support for multi-channel signals.
    """
    batch_size, num_freqs, num_frames = stft_matrix.shape
    device = stft_matrix.device

    if isinstance(analysis_win, np.ndarray):
        analysis_win = torch.tensor(analysis_win, dtype=torch.float32, device=device)

    if isinstance(synthesis_win, np.ndarray):
        synthesis_win = torch.tensor(synthesis_win, dtype=torch.float32, device=device)

    # Initialize the output signal and overlap sum tensors with zeros
    output_length = num_frames * hop_length + win_length
    x = torch.zeros(batch_size, output_length, device=stft_matrix.device)
    win_sum = torch.zeros(batch_size, output_length, device=stft_matrix.device)

    # Pre-compute overlap indices for all frames
    indices = torch.arange(num_frames) * hop_length  # shape: (num_frames,)

    # Compute inverse FFT for all frames at once, utilizing broadcasting
    irfft_frames = torch.fft.irfft(stft_matrix, n=win_length, dim=1)  # shape: (batch_size, win_length, num_frames)

    # Apply window to the results
    for frame_idx in range(num_frames):
        i = indices[frame_idx]

        irfft_frame = irfft_frames[..., frame_idx]  # shape: (batch_size, win_length)
        x[:, i : i + win_length] += irfft_frame * synthesis_win  # Broadcasting

        win_sum[:, i : i + win_length] += analysis_win * synthesis_win  # Broadcasting

    # Avoid division by zero by creating a mask
    pos = win_sum != 0

    # Normalize using the overlap sum
    x[pos] /= win_sum[pos]

    if length is not None:
        # Remove padding of the left (always) and right (if any)
        x = x[:, win_length : win_length + length]

    return x.type(torch.float32)


if __name__ == "__main__":
    # sr: 32 kHz
    # win_length: 512 (16 ms)
    # hop_length: 64 (2 ms)

    win_length = 512
    hop_length = 64
    # sig_length = 1988
    sig_length = 151

    analysis_win = asymmetric_analysis_window(win_length, hop_length)
    synthesis_win = asymmetric_synthesis_window(win_length, hop_length)

    input = torch.rand(2, 2, sig_length)
    torch_frames = asymmetric_stft(input, analysis_win, win_length, hop_length)
    print(torch_frames.shape)

    input = torch.rand(2, sig_length)
    torch_frames = asymmetric_stft(input, analysis_win, win_length, hop_length)
    torch_output = asymmetric_istft(
        torch_frames, analysis_win, synthesis_win, win_length, hop_length, length=sig_length
    )
    print(torch_output.shape)

    print(f"allclose  : {torch.allclose(input[0], torch_output[0])}")
    print(f"mean abs  : {torch.mean(torch.abs(input[0] - torch_output[0]))}")
    print(f"allclose  : {torch.allclose(input[1], torch_output[1])}")
    print(f"mean abs  : {torch.mean(torch.abs(input[1] - torch_output[1]))}")

    print(f"numpy allclose  : {np.allclose(input[0].numpy(), torch_output[0].numpy())}")
    print(f"numpy mean abs  : {np.mean(np.abs(input[0].numpy() - torch_output[0].numpy()))}")



import torch
import torch.nn as nn
from torch.nn.utils.parametrizations import weight_norm

from audiozen.constant import EPSILON


class Chomp1d(nn.Module):
    def __init__(self, chomp_size):
        super(Chomp1d, self).__init__()
        self.chomp_size = chomp_size

    def forward(self, x):
        return x[:, :, : -self.chomp_size].contiguous()


class TemporalBlock(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size, stride, dilation, padding, dropout=0.2):
        super(TemporalBlock, self).__init__()
        self.conv1 = weight_norm(
            nn.Conv1d(
                n_inputs,
                n_outputs,
                kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
            )
        )
        self.chomp1 = Chomp1d(padding)
        self.relu1 = nn.ReLU()
        self.dropout1 = nn.Dropout(dropout)

        self.conv2 = weight_norm(
            nn.Conv1d(
                n_outputs,
                n_outputs,
                kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
            )
        )
        self.chomp2 = Chomp1d(padding)
        self.relu2 = nn.ReLU()
        self.dropout2 = nn.Dropout(dropout)

        self.net = nn.Sequential(
            self.conv1,
            self.chomp1,
            self.relu1,
            self.dropout1,
            self.conv2,
            self.chomp2,
            self.relu2,
            self.dropout2,
        )

        self.downsample = nn.Conv1d(n_inputs, n_outputs, kernel_size=1) if n_inputs != n_outputs else None
        self.relu = nn.ReLU()
        self.init_weights()

    def init_weights(self):
        self.conv1.weight.data.normal_(0, 0.01)
        self.conv2.weight.data.normal_(0, 0.01)
        if self.downsample is not None:
            self.downsample.weight.data.normal_(0, 0.01)

    def forward(self, x):
        out = self.net(x)
        res = x if self.downsample is None else self.downsample(x)
        return self.relu(out + res)


class TemporalConvNet(nn.Module):
    def __init__(self, num_inputs, num_channels, kernel_size=2, dropout=0.2):
        """

        Args:
            num_inputs:
            num_channels: The list of all channels?
            kernel_size:
            dropout:

        Inputs: x
            - x: x has a dimension of [B, C, T]
        """
        super(TemporalConvNet, self).__init__()
        layers = []
        num_levels = len(num_channels)

        for i in range(num_levels):
            dilation_size = 2**i
            in_channels = num_inputs if i == 0 else num_channels[i - 1]
            out_channels = num_channels[i]
            layers += [
                TemporalBlock(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=1,
                    dilation=dilation_size,
                    padding=(kernel_size - 1) * dilation_size,
                    dropout=dropout,
                )
            ]

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)


class CausalConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, encoder_activate_function, **kwargs):
        """

        Args:
            in_channels:
            out_channels:
            encoder_activate_function:
            **kwargs:
        """
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=(3, 2),
            stride=(2, 1),
            padding=(0, 1),
            **kwargs,  # 这里不是左右 pad，而是上下 pad 为 0，左右分别 pad 1...
        )
        self.norm = nn.BatchNorm2d(out_channels)
        self.activation = getattr(nn, encoder_activate_function)()

    def forward(self, x):
        """2D Causal convolution.

        Args:
            x: [B, C, F, T]
        Returns:
            [B, C, F, T]
        """
        x = self.conv(x)
        x = x[:, :, :, :-1]  # chomp size
        x = self.norm(x)
        x = self.activation(x)
        return x


class CausalTransConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, is_last=False, output_padding=(0, 0)):
        super().__init__()
        self.conv = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=(3, 2),
            stride=(2, 1),
            output_padding=output_padding,
        )
        self.norm = nn.BatchNorm2d(out_channels)
        if is_last:
            self.activation = nn.ReLU()
        else:
            self.activation = nn.ELU()

    def forward(self, x):
        """2D Causal convolution.

        Args:
            x: [B, C, F, T]
        Returns:
            [B, C, F, T]
        """
        x = self.conv(x)
        x = x[:, :, :, :-1]  # chomp size
        x = self.norm(x)
        x = self.activation(x)
        return x


class CausalConv2DBlock(nn.Module):
    def __init__(self) -> None:
        super().__init__()

        self.block = nn.Sequential(
            nn.LayerNorm([1, 481, 100]),
            nn.Conv2d(in_channels=1, out_channels=481, kernel_size=1),
        )

    def forward(self, x):
        pass


class TCNBlock(nn.Module):
    def __init__(
        self,
        in_channels=257,
        hidden_channel=512,
        out_channels=257,
        kernel_size=3,
        dilation=1,
        use_skip_connection=True,
        causal=False,
    ):
        super().__init__()
        self.conv1x1 = nn.Conv1d(in_channels, hidden_channel, 1)
        self.prelu1 = nn.PReLU()
        self.norm1 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)
        padding = (dilation * (kernel_size - 1)) // 2 if not causal else (dilation * (kernel_size - 1))
        self.depthwise_conv = nn.Conv1d(
            hidden_channel,
            hidden_channel,
            kernel_size=kernel_size,
            stride=1,
            groups=hidden_channel,
            padding=padding,
            dilation=dilation,
        )
        self.prelu2 = nn.PReLU()
        self.norm2 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)
        self.sconv = nn.Conv1d(hidden_channel, out_channels, 1)
        # self.tcn_block = nn.Sequential(
        #     nn.Conv1d(in_channels, hidden_channel, 1),
        #     nn.PReLU(),
        #     nn.GroupNorm(1, hidden_channel, eps=EPSILON),
        #     nn.Conv1d(hidden_channel, hidden_channel, kernel_size=kernel_size, stride=1,
        #               groups=hidden_channel, padding=padding, dilation=dilation, bias=True),
        #     nn.PReLU(),
        #     nn.GroupNorm(1, hidden_channel, eps=EPSILON),
        #     nn.Conv1d(hidden_channel, out_channels, 1)
        # )

        self.causal = causal
        self.padding = padding
        self.use_skip_connection = use_skip_connection

    def forward(self, x):
        """
        x: [channels, T]
        """
        if self.use_skip_connection:
            y = self.conv1x1(x)
            y = self.norm1(self.prelu1(y))
            y = self.depthwise_conv(y)
            if self.causal:
                y = y[:, :, : -self.padding]
            y = self.norm2(self.prelu2(y))
            output = self.sconv(y)
            return x + output
        else:
            y = self.conv1x1(x)
            y = self.norm1(self.prelu1(y))
            y = self.depthwise_conv(y)
            if self.causal:
                y = y[:, :, : -self.padding]
            y = self.norm2(self.prelu2(y))
            output = self.sconv(y)
            return output


class STCNBlock(nn.Module):
    def __init__(
        self,
        in_channels=257,
        hidden_channel=512,
        out_channels=257,
        kernel_size=3,
        dilation=1,
        use_skip_connection=True,
        causal=False,
    ):
        super().__init__()
        self.conv1x1 = nn.Conv1d(in_channels, hidden_channel, 1)
        self.prelu1 = nn.PReLU()
        self.norm1 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)
        padding = (dilation * (kernel_size - 1)) // 2 if not causal else (dilation * (kernel_size - 1))
        self.depthwise_conv = nn.Conv1d(
            hidden_channel,
            hidden_channel,
            kernel_size=kernel_size,
            stride=1,
            groups=hidden_channel,
            padding=padding,
            dilation=dilation,
        )
        self.prelu2 = nn.PReLU()
        self.norm2 = nn.GroupNorm(1, hidden_channel, eps=EPSILON)
        self.sconv = nn.Conv1d(hidden_channel, out_channels, 1)
        # self.tcn_block = nn.Sequential(
        #     nn.Conv1d(in_channels, hidden_channel, 1),
        #     nn.PReLU(),
        #     nn.GroupNorm(1, hidden_channel, eps=EPSILON),
        #     nn.Conv1d(hidden_channel, hidden_channel, kernel_size=kernel_size, stride=1,
        #               groups=hidden_channel, padding=padding, dilation=dilation, bias=True),
        #     nn.PReLU(),
        #     nn.GroupNorm(1, hidden_channel, eps=EPSILON),
        #     nn.Conv1d(hidden_channel, out_channels, 1)
        # )

        self.causal = causal
        self.padding = padding
        self.use_skip_connection = use_skip_connection

    def forward(self, x):
        """
        x: [channels, T]
        """
        if self.use_skip_connection:
            y = self.conv1x1(x)
            y = self.norm1(self.prelu1(y))
            y = self.depthwise_conv(y)
            if self.causal:
                y = y[:, :, : -self.padding]
            y = self.norm2(self.prelu2(y))
            output = self.sconv(y)
            return x + output
        else:
            y = self.conv1x1(x)
            y = self.norm1(self.prelu1(y))
            y = self.depthwise_conv(y)
            if self.causal:
                y = y[:, :, : -self.padding]
            y = self.norm2(self.prelu2(y))
            output = self.sconv(y)
            return output


if __name__ == "__main__":
    a = torch.rand(2, 1, 19, 200)
    l1 = CausalConvBlock(
        1,
        20,
        kernel_size=(3, 2),
        stride=(2, 1),
        padding=(0, 1),
    )
    l2 = CausalConvBlock(
        20,
        40,
        kernel_size=(3, 2),
        stride=(1, 1),
        padding=1,
    )
    l3 = CausalConvBlock(
        40,
        40,
        kernel_size=(3, 2),
        stride=(2, 1),
        padding=(0, 1),
    )
    l4 = CausalConvBlock(
        40,
        40,
        kernel_size=(3, 2),
        stride=(1, 1),
        padding=1,
    )
    print(l1(a).shape)
    print(l4(l3(l2(l1(a)))).shape)



from functools import partial

import torch
import torch.nn as nn
from einops import rearrange
from torch.nn import functional as F

from audiozen.acoustics.audio_feature import istft, stft
from recipes.intel_ndns.spiking_fullsubnet.efficient_spiking_neuron import MemoryState, efficient_spiking_neuron


class SequenceModel(nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        sequence_model="GSN",
        proj_size=0,
        shared_weights=False,
        output_activate_function=None,
        bn=False,
        use_pre_layer_norm=True,
    ):
        super().__init__()

        if use_pre_layer_norm:
            self.pre_layer_norm = nn.LayerNorm(input_size)

        if sequence_model == "GSN":
            self.sequence_model = efficient_spiking_neuron(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                shared_weights=shared_weights,
                bn=bn,
            )
        elif sequence_model == "LSTM":
            self.sequence_model = nn.LSTM(
                input_size=input_size,
                hidden_size=hidden_size,
                num_layers=num_layers,
                batch_first=True,
                bidirectional=False,
            )
        else:
            raise NotImplementedError(f"Sequence model {sequence_model} not implemented.")

        if proj_size > 0:
            self.proj = nn.Linear(hidden_size, proj_size)
        else:
            self.proj = nn.Identity()

        if output_activate_function == "tanh":
            self.output_activate_function = nn.Tanh()
        elif output_activate_function == "sigmoid":
            self.output_activate_function = nn.Sigmoid()
        elif output_activate_function == "relu":
            self.output_activate_function = nn.ReLU()
        else:
            self.output_activate_function = nn.Identity()

        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.use_pre_layer_norm = use_pre_layer_norm
        self.sequence_model_name = sequence_model

    def forward_lstm(self, input):
        assert input.ndim == 3, f"Input tensor must be 3D, but got {input.ndim}D."
        input = rearrange(input, "b f t -> b t f")

        if self.use_pre_layer_norm:
            input = self.pre_layer_norm(input)

        output, _ = self.sequence_model(input)
        output = self.proj(output)
        output = rearrange(output, "b t f -> b f t")
        output = self.output_activate_function(output)
        return output, []

    def forward(self, input):
        """Forward function.

        Args:
            input (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)` or `(batch_size, num_channels, num_freq, sequence_length)`):
                Input 3D or 4D tensor.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)`): Output tensor.
                Output tensor.
        """
        if self.sequence_model_name == "LSTM":
            return self.forward_lstm(input)

        assert input.ndim == 3, f"Input tensor must be 3D, but got {input.ndim}D."

        batch_size, num_freqs, sequence_length = input.shape

        # Initialize memory states.
        states = [
            MemoryState(
                torch.zeros(batch_size, self.hidden_size, device=input.device),
                torch.zeros(batch_size, self.hidden_size, device=input.device),
            )
            for _ in range(self.num_layers)
        ]

        input = rearrange(input, "b f t -> t b f")

        # Apply layer normalization.
        if self.use_pre_layer_norm:
            input = self.pre_layer_norm(input)

        # Pass through the sequence model.
        output, _, all_layer_outputs = self.sequence_model(input, states)

        # Project the output if necessary.
        output = self.proj(output)
        all_layer_outputs += [output]

        # Apply the output activation function.
        output = self.output_activate_function(output)

        output = rearrange(output, "t b f -> b f t")
        return output, all_layer_outputs


def deepfiltering(complex_spec, coef, order: int, num_spks: int):
    """Deep filtering implementation using `torch.einsum`. Requires unfolded spectrogram.

    Args:
        complex_spec (`torch.ComplexTensor` of shape `[B, C, F, T]`):
            Complex spectrogram.
        coef (`torch.Tensor` of shape `[B, C * order, F, T, 2]`):
            Coefficients of the deep filter.
        order (`int`): Order of the deep filter.
        num_spks (`int`): Number of speakers.

    Returns:
        spec (complex Tensor): Spectrogram of shape `[B, C, S, F, T]`.
    """
    need_unfold = order > 1

    if need_unfold:
        complex_spec = F.pad(complex_spec, (order - 1, 0))
        complex_spec = complex_spec.unfold(3, order, 1)  # [B, C, F, T, df]
    else:
        complex_spec = complex_spec.unsqueeze(-1)  # [B, C, F, T, 1]

    complex_spec = complex_spec.unsqueeze(-1)  # [B, C, F, T, df, 1]
    complex_spec = complex_spec.repeat(1, 1, 1, 1, 1, num_spks)  # [B, C, F, T, df, s]

    complex_coef = torch.complex(coef[..., 0], coef[..., 1])  # [B, C * df, s, F, T]
    complex_coef = rearrange(complex_coef, "b (c df) s f t -> b c df s f t", df=order)

    # [B, C, S, F, T]
    out = torch.einsum("...ftds,...dsft->...sft", complex_spec, complex_coef)

    return out


class Model(nn.Module):
    def __init__(
        self,
        n_fft,
        hop_length,
        win_length,
        fdrc,
        input_size,
        hidden_size,
        num_layers,
        proj_size,
        output_activate_function,
        df_order,
        use_pre_layer_norm_fb=True,
        bn=False,
        shared_weights=False,
        sequence_model="LSTM",
        num_spks=2,
    ):
        super().__init__()

        self.fb_model = SequenceModel(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            shared_weights=shared_weights,
            sequence_model=sequence_model,
            proj_size=proj_size * num_spks * df_order * 2,
            output_activate_function=output_activate_function,
            bn=bn,
            use_pre_layer_norm=use_pre_layer_norm_fb,
        )

        self.stft = partial(stft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.istft = partial(istft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)

        self.fb_input_size = input_size
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        self.fdrc = fdrc
        self.df_order = df_order
        self.num_spks = num_spks

    def forward(self, input):
        """Forward function.

        Args:
            input (`torch.Tensor` of shape `(batch_size, sequence_length)`):
                Waveform tensor.

        Returns:
            output (`torch.Tensor` of shape `(batch_size, num_freq, sequence_length)`):
                Output tensor.
        """
        assert input.ndim == 2, f"Input tensor must be 2D, but got {input.ndim}D."
        batch_size, sequence_length = input.shape

        noisy_mag, _, noisy_real, noisy_imag = self.stft(input)
        noisy_cmp = torch.complex(real=noisy_real, imag=noisy_imag)
        noisy_cmp = rearrange(noisy_cmp, "b f t -> b 1 f t")

        # ================== Fullband model ==================
        noisy_mag = rearrange(noisy_mag, "b f t -> b 1 f t")
        noisy_mag = noisy_mag**self.fdrc
        fb_input = rearrange(noisy_mag, "b c f t -> b (c f) t")
        fb_output, *_ = self.fb_model(fb_input)
        # `c` is corresponding to complex number
        df_coef = rearrange(fb_output, "b (c d s f) t -> b d s f t c", c=2, d=self.df_order, s=self.num_spks)

        # ================== Reconstruct the output ==================
        enh_stft = deepfiltering(noisy_cmp, df_coef, self.df_order, self.num_spks)  # [B, c, s, f, t]

        if self.num_spks > 1:
            enh_stft = rearrange(enh_stft, "b 1 s f t -> (b s) f t")
            enh_y = self.istft(enh_stft, length=sequence_length)
            enh_y = rearrange(enh_y, "(b s) t -> b s t", s=self.num_spks)
            return enh_y, _
        else:
            enh_stft = rearrange(enh_stft, "b 1 1 f t -> b f t")
            enh_mag = torch.abs(enh_stft)  # For computing DNSMOS loss
            enh_y = self.istft(enh_stft, length=sequence_length)
            return enh_y, enh_mag


if __name__ == "__main__":
    model = Model(
        n_fft=512,
        hop_length=128,
        win_length=512,
        fdrc=0.5,
        input_size=257,
        hidden_size=256,
        num_layers=2,
        proj_size=257,
        output_activate_function=None,
        df_order=3,
        use_pre_layer_norm_fb=True,
        bn=False,
        shared_weights=False,
        sequence_model="LSTM",
        num_spks=2,
    )

    input = torch.rand(2, 16000)
    output = model(input)
    print(output[0].shape)



import gc
import importlib
import logging
import os
import random
import time
from typing import Optional

import numpy as np
import torch
from torch import Tensor

from audiozen.accelerate import broadcast_tensor, get_world_size_and_rank


logger = logging.getLogger(__name__)


def check_same_shape(est: Tensor, ref: Tensor) -> None:
    if est.shape != ref.shape:
        raise RuntimeError(f"Dimension mismatch: {est.shape=} vs {ref.shape=}.")


def avoid_nan_grad(optimizer: torch.optim.Optimizer) -> None:
    for group in optimizer.param_groups:
        for p in group["params"]:
            if p.grad is not None and (torch.isnan(p.grad).any() or torch.isinf(p.grad).any()):
                print("NaN or INF found in optimizer gradient")


def avoid_nan_value(tensor: Tensor) -> Tensor:
    if torch.isnan(tensor).any() or torch.isinf(tensor).any():
        logger.warning(f"NaN or INF found in tensor. Shape: {tensor.shape}.")
        tensor[torch.isnan(tensor)] = 0
        tensor[torch.isinf(tensor)] = 0
    return tensor


class Timer:
    """Count execution time.

    Examples:
        >>> timer = ExecutionTime()  # Start timer
        >>> print(f"Finished in {timer.duration()} seconds.")
    """

    def __init__(self):
        self.start_time = time.perf_counter()

    def duration(self, ndigits=3):
        """Get duration of execution.

        Args:
            ndigits: number of digits to round. Default: 3.
        """
        duration = round(time.perf_counter() - self.start_time, ndigits)
        return duration


def log_header(text: str, header_length=78, block_char="="):
    """Print a pretty header in green with a black background"""

    # Limit the length of the text to fit the header
    max_text_length = header_length - 2
    if len(text) > max_text_length:
        text = text[: max_text_length - 3] + "..."
    content = f" {text} "

    logger.info(block_char * header_length)
    # Use center() to automatically pad left and right, aligning the length to header_length
    logger.info(content.center(header_length, block_char))
    logger.info(block_char * header_length)


def instantiate(path: str, args: Optional[dict] = None, initialize: bool = True):
    """Load module or callable (like function) dynamically based on config string.

    Assume that the config items are as follows:

        [model]
            path = "model.FullSubNetModel"
            [model.args]
            n_frames = 32
            ...

    This function will:
        1. Load the "model" module from python search path.
        2. Load "model.FullSubNetModel" class or callable in the "model" module.
        3. If the "initialize" is set to True, instantiate (or call) class (or callable) with args (in "[model.args]").

    Args:
        path: Target class or callable path.
        args: Named arguments passed to class or callable.
        initialize: whether to initialize with args.

    Returns:
        If initialize is True, return the instantiated class or the return of the call.
        Otherwise, return the found class or callable

    Examples:
        >>> # Use official loss function
        >>> instantiate("torch.nn.CrossEntropyLoss", args={"label_smoothing": 0.2}, initialize=True)
        >>> # Use official optimizer
        >>> instantiate("torch.optim.Adam", args={"lr": 1e-3}, initialize=True)
        >>> # Use custom model in a recipe
        >>> instantiate("fsb.model.FullSubNetModel", args={"n_frames": 32}, initialize=True)
        >>> # Use custom loss function in audiozen
        >>> instantiate("audiozen.loss.CRMLoss", initialize=False)
    """
    # e.g., path = "fsb.model.FullSubNetModel"
    # module_path = "fsb.model"
    # class_or_function_name = "FullSubNetModel"
    splitted_path = path.split(".")

    if len(splitted_path) < 2:
        raise ValueError(f"Invalid path: {path}.")

    module_path = ".".join(splitted_path[:-1])
    class_or_function_name = splitted_path[-1]

    module = importlib.import_module(module_path)
    class_or_function = getattr(module, class_or_function_name)

    if initialize:
        if args:
            return class_or_function(**args)
        else:
            return class_or_function()
    else:
        return class_or_function


def set_random_seed(seed: Optional[int] = None):
    """Set random seed for reproducibility.

    Note:
        This function is used to control the reproducibility of the training process.
    """
    world_size, rank = get_world_size_and_rank()
    max_val = np.iinfo(np.uint32).max - world_size + 1
    min_val = np.iinfo(np.uint32).min

    if seed is None:
        rand_seed = torch.randint(min_val, max_val, (1,))
        seed = broadcast_tensor(rand_seed, 0).item()  # sync seed across ranks

    if seed < min_val or seed > max_val:
        raise ValueError(f"Invalid seed value provided: {seed}. Value must be in the range [{min_val}, {max_val}]")

    if rank == 0:
        logger.info(f"Setting manual seed to local seed + rank: {seed + rank}.")

    random.seed(seed + rank)
    np.random.seed(seed + rank)
    torch.manual_seed(seed + rank)


def cleanup_before_training() -> None:
    """Call gc collect, empty CUDA cache, and reset peak memory stats."""
    gc.collect()
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()


def prepare_dirs(dirs):
    """Prepare directories..

    This function creates the specified directories if they don't exist.

    Args:
        dirs: a list of Path objects.
    """
    for dir_path in dirs:
        dir_path.mkdir(parents=True, exist_ok=True)


def prepare_empty_dir(dirs, resume=False):
    """Prepare empty dirs.

    If resume a experiment, this function only assert that dirs should be exist.
    If does not resume a experiment, this function will set up new dirs.

    Args:
        dirs: a list of Path objects.
        resume: whether to resume a experiment. Default: False.
    """
    for dir_path in dirs:
        if resume:
            if not dir_path.exists():
                logger.warning(f"In resume mode, you must have an old experiment directory {dir_path}.")
        else:
            dir_path.mkdir(parents=True, exist_ok=True)


def expand_path(path):
    return os.path.abspath(os.path.expanduser(path))


def clamp_inf_value(tensor):
    """Clamp inf value to a large number."""
    max_dtype = torch.finfo(tensor.dtype).max
    clamp_value = torch.where(torch.isinf(tensor).any(), max_dtype - 1000, tensor)
    tensor = torch.clamp(tensor, min=-clamp_value, max=clamp_value)
    return tensor



from pathlib import Path
from typing import Any, Optional, Tuple, Union

import librosa
import numpy as np
import soundfile as sf
from numpy import ndarray


def load_audio(
    path: Union[Path, str],
    *,
    duration: Optional[float] = None,
    sr: Optional[int] = None,
    **kwargs: Any,
) -> Tuple[np.ndarray, int]:
    """Load the audio using soundfile and resample if necessary.

    Compared to sf.read, this function supports:
    - loading the random segment of a audio when duration is specified.
    - resampling the audio when sr is specified.

    Args:
        path: Path to the audio file.
        duration: Duration of the audio segment in seconds. If None, the whole audio is loaded.
        sr: Sampling rate of the audio. If None, the original sampling rate is used.
        **kwargs: Additional keyword arguments for `np.pad`.

    Returns:
        A tuple of the audio signal and the sampling rate.

    Examples:
        >>> load_audio("test.wav", duration=2.0, sr=16000)
    """
    path = str(path)

    with sf.SoundFile(path) as sf_desc:
        orig_sr = sf_desc.samplerate

        if duration is not None:
            frame_orig_duration = sf_desc.frames
            frame_duration = int(duration * orig_sr)

            if frame_duration < frame_orig_duration:
                # Randomly select a segment
                offset = np.random.randint(frame_orig_duration - frame_duration)
                sf_desc.seek(offset)
                y = sf_desc.read(frames=frame_duration, dtype="float32", always_2d=True).T
            else:
                y = sf_desc.read(dtype="float32", always_2d=True).T  # [C, T]
                if frame_duration > frame_orig_duration:
                    y = np.pad(y, pad_width=((0, 0), (0, frame_duration - frame_orig_duration)), **kwargs)
        else:
            y = sf_desc.read(dtype="float32", always_2d=True).T

    if y.shape[0] == 1:
        y = y.flatten()

    if sr is not None and sr != orig_sr:
        y = librosa.resample(y, orig_sr=orig_sr, target_sr=sr)
        orig_sr = sr

    return y, orig_sr


def extract_segment(data: ndarray, segment_length: int, *, start_idx: int = -1) -> Tuple[ndarray, int]:
    """Sample a segment from the N-dimensional data with the last dimension as the time dimension.

    Args:
        data: Any dimensional data. The last dimension is the time dimension.
        segment_length: The length of the segment to be sampled.
        start_idx: The start index of the segment to be sampled. If less than 0, a random index is generated.
        return_start_idx: Whether to return the start index along with the data segment.

    Returns:
        The sampled data segment and the start index.

    Raises:
        ValueError: If segment_length is negative.
    """
    segment_length = int(segment_length)

    if segment_length < 0:
        raise ValueError("`segment_length` must be non-negative")

    data_len = data.shape[-1]
    ndim = data.ndim

    if data_len > segment_length:
        if start_idx >= 0:
            end_idx = start_idx + segment_length
        else:
            start_idx = np.random.randint(data_len - segment_length)
            end_idx = start_idx + segment_length
        data = data[..., start_idx:end_idx]
    elif data_len < segment_length:
        padding = segment_length - data_len
        pad_width = [(0, 0)] * (ndim - 1) + [(0, padding)]
        data = np.pad(data, pad_width, "constant")
        start_idx = 0

    return data, start_idx



import torch
import torch.nn as nn


class LearnableSigmoid(nn.Module):
    def __init__(self, in_features, beta=1):
        super().__init__()
        self.beta = beta
        self.slope = nn.Parameter(torch.ones(in_features))
        self.slope.requiresGrad = True

    def forward(self, x):
        return self.beta * torch.sigmoid(self.slope * x)


class Discriminator(nn.Module):
    def __init__(self, ndf=16, in_channel=2):
        super().__init__()
        self.layers = nn.Sequential(
            nn.utils.spectral_norm(nn.Conv2d(in_channel, ndf, (4, 4), (2, 2), (1, 1), bias=False)),
            nn.InstanceNorm2d(ndf, affine=True),
            nn.PReLU(ndf),
            nn.utils.spectral_norm(nn.Conv2d(ndf, ndf * 2, (4, 4), (2, 2), (1, 1), bias=False)),
            nn.InstanceNorm2d(ndf * 2, affine=True),
            nn.PReLU(2 * ndf),
            nn.utils.spectral_norm(nn.Conv2d(ndf * 2, ndf * 4, (4, 4), (2, 2), (1, 1), bias=False)),
            nn.InstanceNorm2d(ndf * 4, affine=True),
            nn.PReLU(4 * ndf),
            nn.utils.spectral_norm(nn.Conv2d(ndf * 4, ndf * 8, (4, 4), (2, 2), (1, 1), bias=False)),
            nn.InstanceNorm2d(ndf * 8, affine=True),
            nn.PReLU(8 * ndf),
            nn.AdaptiveMaxPool2d(1),
            nn.Flatten(),
            nn.utils.spectral_norm(nn.Linear(ndf * 8, ndf * 4)),
            nn.Dropout(0.3),
            nn.PReLU(4 * ndf),
            nn.utils.spectral_norm(nn.Linear(ndf * 4, 1)),
            LearnableSigmoid(1),
        )

    def forward(self, x, y):
        if x.dim() == 3:
            x = x.unsqueeze(1)

        if y.dim() == 3:
            y = y.unsqueeze(1)

        xy = torch.cat([x, y], dim=1)
        return self.layers(xy)


if __name__ == "__main__":
    model = Discriminator()

    x = torch.rand(1, 1, 257, 32000 // 128 + 1)
    y = torch.rand(1, 1, 257, 32000 // 128 + 1)

    print(model(x, y).shape, model(x, y))



import math
from functools import partial
from typing import Optional, Tuple, Union

import torch
from einops import rearrange
from torch import nn

from audiozen.acoustics.audio_feature import istft, stft
from audiozen.models.transformer.configuration_transformer import ModelArgs
from audiozen.models.transformer.lora import LoRALinear


def maybe_lora(args: ModelArgs) -> Union[nn.Linear, LoRALinear]:
    if args.lora is None:
        return nn.Linear
    else:
        return partial(LoRALinear, rank=args.lora.rank, scaling=args.lora.scaling)


class PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[seq_len, batch_size, embedding_dim]``
        """
        x = x + self.pe[: x.size(0)]
        return self.dropout(x)


def precompute_freqs_cis(dim: int, end: int, theta: float) -> torch.Tensor:
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    t = torch.arange(end, device=freqs.device)
    freqs = torch.outer(t, freqs).float()
    out = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return out


def apply_rotary_emb(xq: torch.Tensor, xk: torch.Tensor, freqs_cis: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
    freqs_cis = freqs_cis[:, None, :]
    # original code is 2 but it should be 3 as the dim of batch is missing
    xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(3)
    xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)


class FeedForward(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()

        MaybeLora = maybe_lora(args)
        self.w1 = MaybeLora(args.dim, args.hidden_dim, bias=False)
        self.w2 = MaybeLora(args.hidden_dim, args.dim, bias=False)
        self.w3 = MaybeLora(args.dim, args.hidden_dim, bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.w2(nn.functional.silu(self.w1(x)) * self.w3(x))


class RMSNorm(torch.nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


class Attention(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.args = args

        self.n_heads: int = args.n_heads
        self.head_dim: int = args.head_dim

        self.scale = self.args.head_dim**-0.5

        MaybeLora = maybe_lora(args)
        self.wq = MaybeLora(args.dim, args.n_heads * args.head_dim, bias=False)
        self.wk = MaybeLora(args.dim, args.n_heads * args.head_dim, bias=False)
        self.wv = MaybeLora(args.dim, args.n_heads * args.head_dim, bias=False)
        self.wo = MaybeLora(args.n_heads * args.head_dim, args.dim, bias=False)

        self.positional_embedding = PositionalEncoding(args.dim)

    def forward(self, x: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
        # x: [batch_size, seqlen, dim]
        # output: [batch_size, seqlen, dim]
        # N: number of heads
        # D: head_dim
        # T: seqlen

        # We are now having an input with the shape of [B, T, E]
        # Now, we are using the linear layers to project the input into the query, key, and value with the shape of [B, T, N * D]
        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

        # We are now splitting the query, key, and value into N heads with the shape of [B, T, N, D]
        xq = rearrange(xq, "b t (n d) -> b t n d", n=self.n_heads)
        xk = rearrange(xk, "b t (n d) -> b t n d", n=self.n_heads)
        xv = rearrange(xv, "b t (n d) -> b t n d", n=self.n_heads)

        # apply_rotary_emb
        xq, xk = apply_rotary_emb(xq, xk, freqs_cis)

        xq = rearrange(xq, "b t n d -> b n t d", n=self.n_heads)
        xk = rearrange(xk, "b t n d -> b n t d", n=self.n_heads)
        xv = rearrange(xv, "b t n d -> b n t d", n=self.n_heads)

        output = nn.functional.scaled_dot_product_attention(xq, xk, xv, scale=self.scale)
        output = rearrange(output, "b n t d -> b t (n d)")
        return self.wo(output)


class TransformerLayer(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.n_heads = args.n_heads
        self.dim = args.dim
        self.attention = Attention(args)
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.args = args

        self.feed_forward: nn.Module
        self.feed_forward = FeedForward(args=args)

    def forward(self, x: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
        r = self.attention(self.attention_norm(x), freqs_cis)
        h = x + r
        r = self.feed_forward(self.ffn_norm(h))
        out = h + r
        return out


class TransformerEncoder(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.layers = nn.ModuleList([TransformerLayer(args) for _ in range(args.n_layers)])

    def forward(self, x: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
        for layer in self.layers:
            x = layer(x, freqs_cis)
        return x


class TransformerEncoderWrapper(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.args = args
        self.encoder = TransformerEncoder(args)

        self._precomputed_freqs_cis: Optional[torch.Tensor] = None

        self.stft = partial(stft, n_fft=512, hop_length=128, win_length=512)
        self.istft = partial(istft, n_fft=512, hop_length=128, win_length=512)

    @property
    def device(self) -> torch.device:
        return next(self.parameters()).device

    @property
    def freqs_cis(self) -> torch.Tensor:
        # We cache freqs_cis but need to take care that it is on the right device
        # and has the right dtype (complex64). The fact that the dtype is different
        # from the module's  dtype means we cannot register it as a buffer
        if self._precomputed_freqs_cis is None:
            # default to 10**6
            theta = self.args.rope_theta or 1000000.0
            self._precomputed_freqs_cis = precompute_freqs_cis(self.args.head_dim, 128_000, theta)

        if self._precomputed_freqs_cis.device != self.device:
            self._precomputed_freqs_cis = self._precomputed_freqs_cis.to(device=self.device)

        return self._precomputed_freqs_cis

    def forward(self, noisy_y: torch.Tensor) -> torch.Tensor:
        # noisy_y: [B, C, T] or [B, T]
        # enh_y: [B, T]
        if noisy_y.dim() == 3:
            assert noisy_y.size(1) == 1, "Only mono audio is supported."
            noisy_y = noisy_y.squeeze(1)

        num_samples = noisy_y.size(-1)
        mag_mix, phase_mix, real_mix, imag_mix = self.stft(noisy_y)

        # seq_len
        positions = torch.arange(0, mag_mix.shape[-1]).to(device=mag_mix.device, dtype=torch.long)
        freqs_cis = self.freqs_cis[positions]

        # remove direct component
        real_mix = real_mix[:, 1:]
        imag_mix = imag_mix[:, 1:]

        feat = torch.cat([real_mix, imag_mix], dim=1)  # [B, 2 * F, T]

        feat = rearrange(feat, "b f t -> b t f")
        mask = self.encoder(feat, freqs_cis)
        mask = rearrange(mask, "b t f -> b f t")
        mask_real = mask[:, :256]
        mask_imag = mask[:, 256:]

        enh_real = real_mix * mask_real - imag_mix * mask_imag
        enh_imag = real_mix * mask_imag + imag_mix * mask_real

        # add direct component
        enh_real = torch.cat([real_mix[:, 0:1], enh_real], dim=1)  # [B, F, T]
        enh_imag = torch.cat([imag_mix[:, 0:1], enh_imag], dim=1)

        enh = torch.complex(enh_real, enh_imag)

        enh = self.istft(enh, length=num_samples, input_type="complex")

        return enh


if __name__ == "__main__":
    args = ModelArgs(dim=512, n_layers=6, head_dim=256, hidden_dim=2048, n_heads=8, norm_eps=1e-6)
    model = TransformerEncoderWrapper(args)
    output = model(torch.rand(2, 1, 16000))
    print(output.shape)



import torch
import torch.nn as nn


class GGRU(nn.Module):
    def __init__(
        self,
        in_features=None,
        out_features=None,
        mid_features=None,
        hidden_size=1024,
        groups=2,
    ):
        super(GGRU, self).__init__()
        hidden_size_t = hidden_size // groups
        self.gru_list1 = nn.ModuleList(
            [nn.GRU(hidden_size_t, hidden_size_t, 1, batch_first=True) for i in range(groups)]
        )

        self.gru_list2 = nn.ModuleList(
            [nn.GRU(hidden_size_t, hidden_size_t, 1, batch_first=True) for i in range(groups)]
        )
        self.ln1 = nn.LayerNorm(hidden_size)
        self.ln2 = nn.LayerNorm(hidden_size)
        self.groups = groups
        self.mid_features = mid_features

    def forward(self, x):
        out = x
        out = out.transpose(1, 2).contiguous()
        out = out.view(out.size(0), out.size(1), -1).contiguous()

        out = torch.chunk(out, self.groups, dim=-1)
        out = torch.stack([self.gru_list1[i](out[i])[0] for i in range(self.groups)], dim=-1)
        out = torch.flatten(out, start_dim=-2, end_dim=-1)
        out = self.ln1(out)

        out = torch.chunk(out, self.groups, dim=-1)
        out = torch.cat([self.gru_list2[i](out[i])[0] for i in range(self.groups)], dim=-1)
        out = self.ln2(out)

        out = self.view(out.size(0), out.size(1), x.size(1), -1).contiguous()
        out = out.transpose(1, 2).contiguous()
        return out


class unet1(nn.Module):
    def __init__(
        self,
        in_features,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
    ) -> None:
        super(unet1, self).__init__()
        self.in_features = in_features
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride
        self.kernel_size = kernel_size
        self.padding = padding

        self.conv1 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            stride=stride,
            kernel_size=kernel_size,
        )

        self.conv2 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )

        self.conv3 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )

        self.gru = GGRU(groups=2)

        self.conv2_t_1 = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )

        self.conv1_t_1 = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.fc = nn.Linear(in_features=in_features, out_features=in_features)

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.bn2 = nn.BatchNorm2d(in_channels)
        self.bn3 = nn.BatchNorm2d(in_channels)
        self.bn2_t_1 = nn.BatchNorm2d(in_channels)
        self.bn1_t_1 = nn.BatchNorm2d(in_channels)
        self.elu = nn.ELU(inplace=True)

    def forward(self, x):
        out = x
        e1 = self.elu(self.bn1(self.conv1(out)))
        e2 = self.elu(self.bn2(self.conv2(e1)))
        e3 = self.elu(self.bn3(self.conv3(e2)))

        out = e3
        out = self.gru(out)
        out1 = self.elu(self.bn2_t_1(self.conv2_t_1(out)))
        out1 = self.elu(self.bn1_t_1(self.conv2_t_1(out1)))
        out1 = self.fc(out1)

        return out


if __name__ == "__main__":
    x = torch.randn((2, 1, 10, 161))
    net = unet1()
    y = net(x)
    print(f"{x.shape}->{y.shape}")



import logging
from pathlib import Path
from typing import Union

import librosa
import numpy as np
import onnxruntime as ort
import torch
from pesq import pesq as pesq_backend
from pystoi import stoi as stoi_backend

from audiozen.utils import check_same_shape


ort.preload_dlls()

options = ort.SessionOptions()
options.intra_op_num_threads = 1
options.inter_op_num_threads = 1

logger = logging.getLogger(__name__)


def preprocessing(est, ref):
    if est.ndim != 1 or ref.ndim != 1:
        est = est.reshape(-1)
        ref = ref.reshape(-1)

    check_same_shape(est, ref)

    if torch.is_tensor(est) or torch.is_tensor(ref):
        est = est.detach().cpu().numpy()
        ref = ref.detach().cpu().numpy()

    return est, ref


class STOI:
    def __init__(self, sr=16000) -> None:
        self.sr = sr

    def __call__(self, est, ref, extended=False):
        est, ref = preprocessing(est, ref)
        stoi_val = stoi_backend(ref, est, self.sr, extended=extended)
        return {"stoi": stoi_val}


class PESQ:
    def __init__(self, sr=16000, mode="wb") -> None:
        self.sr = sr
        self.mode = mode

        if self.sr not in (8000, 16000):
            logging.warning(
                f"Unsupported sample rate: {sr}. Expected 8000 or 16000. Resampling will be applied later."
            )

        if self.mode not in ("wb", "nb"):
            raise ValueError(f"Unsupported mode: {self.mode}. Expected 'wb' or 'nb'.")

    def __call__(self, est, ref):
        est, ref = preprocessing(est, ref)
        if self.sr not in (8000, 16000):
            ref = librosa.resample(ref, orig_sr=self.sr, target_sr=16000)
            est = librosa.resample(est, orig_sr=self.sr, target_sr=16000)
            self.sr = 16000

        pesq_val = pesq_backend(self.sr, ref, est, self.mode)

        return {f"pesq_{self.mode}": pesq_val}


class SISDR:
    def __init__(self) -> None:
        super().__init__()

    def __call__(
        self,
        estimate: Union[torch.Tensor, np.ndarray],
        target: Union[torch.Tensor, np.ndarray],
        reduce_mean: bool = True,
    ) -> dict[str, float]:
        if not torch.is_tensor(target):
            target = torch.tensor(target)
        if not torch.is_tensor(estimate):
            estimate = torch.tensor(estimate)

        eps = torch.finfo(estimate.dtype).eps

        # zero mean to ensure scale invariance
        s_target = target - torch.mean(target, dim=-1, keepdim=True)
        s_estimate = estimate - torch.mean(estimate, dim=-1, keepdim=True)

        # <s, s'> / ||s||**2 * s
        pair_wise_dot = torch.sum(s_target * s_estimate, dim=-1, keepdim=True)
        s_target_norm = torch.sum(s_target**2, dim=-1, keepdim=True)
        pair_wise_proj = (pair_wise_dot * s_target + eps) / (s_target_norm + eps)

        e_noise = s_estimate - pair_wise_proj

        pair_wise_sdr = (torch.sum(pair_wise_proj**2, dim=-1) + eps) / (torch.sum(e_noise**2, dim=-1) + eps)
        val = 10 * torch.log10(pair_wise_sdr + eps)

        if reduce_mean:
            val = torch.mean(val)

        return {"si_sdr": val.item()}


class pDNSMOS:
    def __init__(self, input_sr=16000) -> None:
        super().__init__()

        root_dir = Path(__file__).parent.absolute()

        self.p835_personal_sess = ort.InferenceSession(
            root_dir / "external" / "pDNSMOS" / "sig_bak_ovr.onnx",
            providers=["CPUExecutionProvider"],
        )

        self.input_sr = input_sr

    def audio_melspec(self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True):
        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40  # type: ignore
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_MOS=False):
        if is_personalized_MOS:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(self, audio):
        if audio.ndim != 1:
            audio = audio.reshape(-1)

        if torch.is_tensor(audio):
            audio = audio.detach().cpu().numpy()

        SAMPLERATE = 16000
        INPUT_LENGTH = 9.01

        if self.input_sr != 16000:
            audio = librosa.resample(audio, orig_sr=self.input_sr, target_sr=SAMPLERATE)

        len_samples = int(INPUT_LENGTH * SAMPLERATE)

        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / SAMPLERATE) - INPUT_LENGTH) + 1

        hop_len_samples = SAMPLERATE
        predicted_p_mos_sig_seg_raw = []
        predicted_p_mos_bak_seg_raw = []
        predicted_p_mos_ovr_seg_raw = []

        for idx in range(num_hops):
            audio_seg = audio[int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            # p808_input_features = np.array(self.audio_melspec(audio=audio_seg[:-160])).astype("float32")[
            # np.newaxis, :, :
            # ]
            oi = {"input_1": input_features}
            # p808_oi = {"input_1": p808_input_features}
            p_mos_sig_raw, p_mos_bak_raw, p_mos_ovr_raw = self.p835_personal_sess.run(None, oi)[0][0]

            predicted_p_mos_ovr_seg_raw.append(p_mos_ovr_raw)
            predicted_p_mos_sig_seg_raw.append(p_mos_sig_raw)
            predicted_p_mos_bak_seg_raw.append(p_mos_bak_raw)

        clip_dict = {}
        # clip_dict["sr"] = SAMPLERATE
        # clip_dict["len"] = actual_audio_len / SAMPLERATE
        clip_dict["pSIG"] = np.mean(predicted_p_mos_sig_seg_raw)
        clip_dict["pBAK"] = np.mean(predicted_p_mos_bak_seg_raw)
        clip_dict["pOVRL"] = np.mean(predicted_p_mos_ovr_seg_raw)

        return clip_dict


class DNSMOS:
    def __init__(self, input_sr=16000, device=-1, use_p808=True) -> None:
        super().__init__()

        root_dir = Path(__file__).parent.absolute()

        # =================== Hack to fix the issue with onnxruntime ===================
        # https://github.com/microsoft/onnxruntime/issues/8313#issuecomment-1486097717
        # Globally modify the default inter_op_num_threads and intra_op_num_threads parameters
        # (Only needed for Slurm system as it has a cgroup)
        # _default_session_options = ort.capi._pybind_state.get_default_session_options()

        # def get_default_session_options_new():
        #     _default_session_options.inter_op_num_threads = 1
        #     _default_session_options.intra_op_num_threads = 1
        #     return _default_session_options

        # ort.capi._pybind_state.get_default_session_options = get_default_session_options_new
        # =================== Hack to fix the issue with onnxruntime ===================

        if device > 0:
            logger.info(f"[DNSMOS] Using CUDA Provider with device_id: {device}")
            providers = [("CUDAExecutionProvider", {"device_id": device})]
        else:
            providers = ["CPUExecutionProvider"]

        self.p835_sess = ort.InferenceSession(
            (root_dir / "external" / "DNSMOS" / "sig_bak_ovr.onnx").as_posix(),
            providers=providers,
        )

        if use_p808:
            self.p808_sess = ort.InferenceSession(
                (root_dir / "external" / "DNSMOS" / "model_v8.onnx").as_posix(),
                providers=providers,
            )

        self.input_sr = input_sr
        self.use_p808 = use_p808

    def audio_melspec(self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True):
        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_MOS=False):
        if is_personalized_MOS:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(self, audio):
        if audio.ndim != 1:
            audio = audio.reshape(-1)

        if torch.is_tensor(audio):
            audio = audio.detach().cpu().numpy()

        SAMPLERATE = 16000
        INPUT_LENGTH = 9.01

        if self.input_sr != 16000:
            audio = librosa.resample(audio, orig_sr=self.input_sr, target_sr=SAMPLERATE)

        len_samples = int(INPUT_LENGTH * SAMPLERATE)

        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / SAMPLERATE) - INPUT_LENGTH) + 1

        hop_len_samples = SAMPLERATE
        predicted_mos_sig_seg = []
        predicted_mos_bak_seg = []
        predicted_mos_ovr_seg = []
        predicted_p808_mos = []

        for idx in range(num_hops):
            audio_seg = audio[int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            oi = {"input_1": input_features}

            if self.use_p808:
                p808_input_features = np.array(self.audio_melspec(audio=audio_seg[:-160])).astype("float32")[
                    np.newaxis, :, :
                ]
                p808_oi = {"input_1": p808_input_features}
                p808_mos = self.p808_sess.run(None, p808_oi)[0][0][0]
                predicted_p808_mos.append(p808_mos)

            mos_sig_raw, mos_bak_raw, mos_ovr_raw = self.p835_sess.run(None, oi)[0][0]
            mos_sig, mos_bak, mos_ovr = self.get_polyfit_val(mos_sig_raw, mos_bak_raw, mos_ovr_raw)

            predicted_mos_ovr_seg.append(mos_ovr)
            predicted_mos_sig_seg.append(mos_sig)
            predicted_mos_bak_seg.append(mos_bak)

        clip_dict = {}

        if self.use_p808:
            clip_dict["P808"] = np.mean(predicted_p808_mos)

        clip_dict["OVRL"] = np.mean(predicted_mos_ovr_seg)
        clip_dict["SIG"] = np.mean(predicted_mos_sig_seg)
        clip_dict["BAK"] = np.mean(predicted_mos_bak_seg)

        return clip_dict


def compute_synops(fb_all_layer_outputs, sb_all_layer_outputs, shared_weights=True):
    synops = 0.0
    for i in range(1, len(fb_all_layer_outputs) - 1):
        curr_synops = (
            torch.gt(fb_all_layer_outputs[i], 0).float().mean()
            * fb_all_layer_outputs[i].size(-1)
            * (fb_all_layer_outputs[i + 1].size(-1) + fb_all_layer_outputs[i].size(-1))
        )
        # print(f"i: {i}, synops: {curr_synops}")
        synops += curr_synops
    for i in range(len(sb_all_layer_outputs)):
        for j in range(1, len(sb_all_layer_outputs[i]) - 1):
            # print(sb_all_layer_outputs[i][j].size())
            curr_synops = (
                torch.gt(sb_all_layer_outputs[i][j], 0).float().mean()
                * sb_all_layer_outputs[i][j].size(-1)
                * (sb_all_layer_outputs[i][j + 1].size(-1) + sb_all_layer_outputs[i][j].size(-1))
            )
            # print(f"i: {i}, j: {j}, synops: {curr_synops}")
            synops += curr_synops

    if shared_weights:
        return synops.item()
    else:
        return 2 * synops.item()


def compute_synops_v2(fb_all_layer_outputs, sb_all_layer_outputs, shared_weights=True):
    synops = 0.0
    T = fb_all_layer_outputs[1].size(0)
    batch_size = fb_all_layer_outputs[1].size(1)
    for i in range(1, len(fb_all_layer_outputs) - 1):
        curr_synops = (
            torch.gt(fb_all_layer_outputs[i], 0).float().sum()
            / (T * batch_size)
            * (fb_all_layer_outputs[i + 1].size(-1) + fb_all_layer_outputs[i].size(-1))
        )
        print(f"i: {i}, synops: {curr_synops}")
        synops += curr_synops
    for i in range(len(sb_all_layer_outputs)):
        for j in range(1, len(sb_all_layer_outputs[i]) - 1):
            # print(sb_all_layer_outputs[i][j].size())
            curr_synops = (
                torch.gt(sb_all_layer_outputs[i][j], 0).float().sum()
                / (T * batch_size)
                * (sb_all_layer_outputs[i][j + 1].size(-1) + sb_all_layer_outputs[i][j].size(-1))
            )
            print(f"i: {i}, j: {j}, synops: {curr_synops}")
            synops += curr_synops

    if shared_weights:
        return synops.item()
    else:
        return 2 * synops.item()


def compute_neuronops(fb_all_layer_outputs, sb_all_layer_outputs):
    # print("Compute neuronops.....")
    neuronops = 0.0
    for i in range(len(fb_all_layer_outputs)):
        # print(f"i:{i}, {fb_all_layer_outputs[i].size()}")
        neuronops += fb_all_layer_outputs[i].size(-1)
    for i in range(len(sb_all_layer_outputs)):
        for j in range(len(sb_all_layer_outputs[i])):
            # print(f"i:{i} j: {j}, {sb_all_layer_outputs[i][j].size()}")
            neuronops += sb_all_layer_outputs[i][j].size(-1)
    return neuronops


if __name__ == "__main__":
    loss = SISDR()
    est = torch.randn(2, 2, 16000)
    ref = torch.randn(2, 2, 16000)
    print(loss(est, ref))



import torch
import torch.nn as nn
from einops import pack, rearrange, reduce, repeat
from torch import Tensor

from audiozen.constant import EPSILON


class TransformAverageConcatenate(nn.Module):
    def __init__(self, input_size, hidden_size) -> None:
        super().__init__()

        self.transform_layer = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.GELU(),
        )
        self.average_layer = nn.Sequential(
            nn.Linear(hidden_size, hidden_size),
            nn.GELU(),
        )
        self.output_layer = nn.Sequential(
            nn.Linear(hidden_size * 2, input_size),
            nn.GELU(),
        )

        self.norm_layer = nn.GroupNorm(1, input_size, eps=EPSILON)

    def forward(self, input: Tensor):
        batch_size, num_groups, num_freqs, num_frames = input.shape

        output = rearrange(input, "B G F T -> (B G) F T")
        output = self.norm_layer(output)
        output = rearrange(output, "(B G) F T -> B G F T", B=batch_size)

        # Transform
        group_input = rearrange(output, "B G F T -> (B G T) F")
        group_output = self.transform_layer(group_input)
        group_output = rearrange(group_output, "(B G T) F -> B T G F", B=batch_size, G=num_groups)

        # Average
        group_mean = reduce(group_output, "B T G F -> B T F", "mean")
        group_mean = rearrange(group_mean, "B T F -> (B T) F")
        group_mean = self.average_layer(group_mean)
        group_mean = repeat(group_mean, "BT F -> BT G F", G=num_groups)

        # Concatenate
        group_output = rearrange(group_output, "B T G F -> (B T) G F")
        group_output, _ = pack([group_output, group_mean], "BT G *")
        group_output = self.output_layer(group_output)  # [BT, G, F]
        group_output = rearrange(group_output, "(B T) G F -> B G F T", B=batch_size)

        return input + group_output


if __name__ == "__main__":
    input = torch.rand(2, 4, 64, 256)
    model = TransformAverageConcatenate(64, 256)
    output = model(input)
    print(output.shape)



"""Implements Conv-TasNet with building blocks of it.

Copy from torchaudio.models.conv_tasnet
"""

from typing import Optional, Tuple

import torch


class ConvBlock(torch.nn.Module):
    """1D Convolutional block.

    Args:
        io_channels (int): The number of input/output channels, <B, Sc>
        hidden_channels (int): The number of channels in the internal layers, <H>.
        kernel_size (int): The convolution kernel size of the middle layer, <P>.
        padding (int): Padding value of the convolution in the middle layer.
        dilation (int, optional): Dilation value of the convolution in the middle layer.
        no_redisual (bool, optional): Disable residual block/output.

    Note:
        This implementation corresponds to the "non-causal" setting in the paper.
    """

    def __init__(
        self,
        io_channels: int,
        hidden_channels: int,
        kernel_size: int,
        padding: int,
        dilation: int = 1,
        no_residual: bool = False,
    ):
        super().__init__()

        self.conv_layers = torch.nn.Sequential(
            torch.nn.Conv1d(in_channels=io_channels, out_channels=hidden_channels, kernel_size=1),
            torch.nn.PReLU(),
            torch.nn.GroupNorm(num_groups=1, num_channels=hidden_channels, eps=1e-08),
            torch.nn.Conv1d(
                in_channels=hidden_channels,
                out_channels=hidden_channels,
                kernel_size=kernel_size,
                padding=padding,
                dilation=dilation,
                groups=hidden_channels,
            ),
            torch.nn.PReLU(),
            torch.nn.GroupNorm(num_groups=1, num_channels=hidden_channels, eps=1e-08),
        )

        self.res_out = (
            None
            if no_residual
            else torch.nn.Conv1d(in_channels=hidden_channels, out_channels=io_channels, kernel_size=1)
        )
        self.skip_out = torch.nn.Conv1d(in_channels=hidden_channels, out_channels=io_channels, kernel_size=1)

    def forward(self, input: torch.Tensor) -> Tuple[Optional[torch.Tensor], torch.Tensor]:
        feature = self.conv_layers(input)
        if self.res_out is None:
            residual = None
        else:
            residual = self.res_out(feature)
        skip_out = self.skip_out(feature)
        return residual, skip_out


class MaskGenerator(torch.nn.Module):
    """TCN (Temporal Convolution Network) Separation Module.

    Generates masks for separation.

    Args:
        input_dim (int): Input feature dimension, <N>.
        num_sources (int): The number of sources to separate.
        kernel_size (int): The convolution kernel size of conv blocks, <P>.
        num_featrs (int): Input/output feature dimenstion of conv blocks, <B, Sc>.
        num_hidden (int): Intermediate feature dimention of conv blocks, <H>
        num_layers (int): The number of conv blocks in one stack, <X>.
        num_stacks (int): The number of conv block stacks, <R>.
        msk_activate (str): The activation function of the mask output.

    Note:
        This implementation corresponds to the "non-causal" setting in the paper.
    """

    def __init__(
        self,
        input_dim: int,
        num_sources: int,
        kernel_size: int,
        num_feats: int,
        num_hidden: int,
        num_layers: int,
        num_stacks: int,
        msk_activate: str,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.num_sources = num_sources

        self.input_norm = torch.nn.GroupNorm(num_groups=1, num_channels=input_dim, eps=1e-8)
        self.input_conv = torch.nn.Conv1d(in_channels=input_dim, out_channels=num_feats, kernel_size=1)

        self.receptive_field = 0
        self.conv_layers = torch.nn.ModuleList([])
        for s in range(num_stacks):
            for l in range(num_layers):
                multi = 2**l
                self.conv_layers.append(
                    ConvBlock(
                        io_channels=num_feats,
                        hidden_channels=num_hidden,
                        kernel_size=kernel_size,
                        dilation=multi,
                        padding=multi,
                        # The last ConvBlock does not need residual
                        no_residual=(l == (num_layers - 1) and s == (num_stacks - 1)),
                    )
                )
                self.receptive_field += kernel_size if s == 0 and l == 0 else (kernel_size - 1) * multi
        self.output_prelu = torch.nn.PReLU()
        self.output_conv = torch.nn.Conv1d(
            in_channels=num_feats,
            out_channels=input_dim * num_sources,
            kernel_size=1,
        )
        if msk_activate == "sigmoid":
            self.mask_activate = torch.nn.Sigmoid()
        elif msk_activate == "relu":
            self.mask_activate = torch.nn.ReLU()
        else:
            raise ValueError(f"Unsupported activation {msk_activate}")

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Generate separation mask.

        Args:
            input (torch.Tensor): 3D Tensor with shape [batch, features, frames]

        Returns:
            Tensor: shape [batch, num_sources, features, frames]
        """
        batch_size = input.shape[0]
        feats = self.input_norm(input)
        feats = self.input_conv(feats)
        output = 0.0
        for layer in self.conv_layers:
            residual, skip = layer(feats)
            if residual is not None:  # the last conv layer does not produce residual
                feats = feats + residual
            output = output + skip
        output = self.output_prelu(output)
        output = self.output_conv(output)
        output = self.mask_activate(output)
        return output.view(batch_size, self.num_sources, self.input_dim, -1)


class ConvTasNet(torch.nn.Module):
    """Conv-TasNet architecture introduced in *Conv-TasNet: Surpassing Ideal Time–Frequency Magnitude Masking for
    Speech Separation* :cite:`Luo_2019`.

    Note:
        This implementation corresponds to the "non-causal" setting in the paper.

    See Also:
        * :class:`torchaudio.pipelines.SourceSeparationBundle`: Source separation pipeline with pre-trained models.

    Args:
        num_sources (int, optional): The number of sources to split.
        enc_kernel_size (int, optional): The convolution kernel size of the encoder/decoder, <L>.
        enc_num_feats (int, optional): The feature dimensions passed to mask generator, <N>.
        msk_kernel_size (int, optional): The convolution kernel size of the mask generator, <P>.
        msk_num_feats (int, optional): The input/output feature dimension of conv block in the mask generator, <B, Sc>.
        msk_num_hidden_feats (int, optional): The internal feature dimension of conv block of the mask generator, <H>.
        msk_num_layers (int, optional): The number of layers in one conv block of the mask generator, <X>.
        msk_num_stacks (int, optional): The numbr of conv blocks of the mask generator, <R>.
        msk_activate (str, optional): The activation function of the mask output (Default: ``sigmoid``).
    """

    def __init__(
        self,
        num_sources: int = 2,
        # encoder/decoder parameters
        enc_kernel_size: int = 16,
        enc_num_feats: int = 512,
        # mask generator parameters
        msk_kernel_size: int = 3,
        msk_num_feats: int = 128,
        msk_num_hidden_feats: int = 512,
        msk_num_layers: int = 8,
        msk_num_stacks: int = 3,
        msk_activate: str = "sigmoid",
    ):
        super().__init__()

        self.num_sources = num_sources
        self.enc_num_feats = enc_num_feats
        self.enc_kernel_size = enc_kernel_size
        self.enc_stride = enc_kernel_size // 2

        self.encoder = torch.nn.Conv1d(
            in_channels=1,
            out_channels=enc_num_feats,
            kernel_size=enc_kernel_size,
            stride=self.enc_stride,
            padding=self.enc_stride,
            bias=False,
        )
        self.mask_generator = MaskGenerator(
            input_dim=enc_num_feats,
            num_sources=num_sources,
            kernel_size=msk_kernel_size,
            num_feats=msk_num_feats,
            num_hidden=msk_num_hidden_feats,
            num_layers=msk_num_layers,
            num_stacks=msk_num_stacks,
            msk_activate=msk_activate,
        )
        self.decoder = torch.nn.ConvTranspose1d(
            in_channels=enc_num_feats,
            out_channels=1,
            kernel_size=enc_kernel_size,
            stride=self.enc_stride,
            padding=self.enc_stride,
            bias=False,
        )

    def _align_num_frames_with_strides(self, input: torch.Tensor) -> Tuple[torch.Tensor, int]:
        """Pad input Tensor so that the end of the input tensor corresponds with.

        1. (if kernel size is odd) the center of the last convolution kernel
        or 2. (if kernel size is even) the end of the first half of the last convolution kernel

        Assumption:
            The resulting Tensor will be padded with the size of stride (== kernel_width // 2)
            on the both ends in Conv1D

        |<--- k_1 --->|
        |      |            |<-- k_n-1 -->|
        |      |                  |  |<--- k_n --->|
        |      |                  |         |      |
        |      |                  |         |      |
        |      v                  v         v      |
        |<---->|<--- input signal --->|<--->|<---->|
         stride                         PAD  stride

        Args:
            input (torch.Tensor): 3D Tensor with shape (batch_size, channels==1, frames)

        Returns:
            Tensor: Padded Tensor
            int: Number of paddings performed
        """
        batch_size, num_channels, num_frames = input.shape
        is_odd = self.enc_kernel_size % 2
        num_strides = (num_frames - is_odd) // self.enc_stride
        num_remainings = num_frames - (is_odd + num_strides * self.enc_stride)
        if num_remainings == 0:
            return input, 0

        num_paddings = self.enc_stride - num_remainings
        pad = torch.zeros(
            batch_size,
            num_channels,
            num_paddings,
            dtype=input.dtype,
            device=input.device,
        )
        return torch.cat([input, pad], 2), num_paddings

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Perform source separation. Generate audio source waveforms.

        Args:
            input (torch.Tensor): 3D Tensor with shape [batch, channel==1, frames]

        Returns:
            Tensor: 3D Tensor with shape [batch, channel==num_sources, frames]
        """
        input = input.unsqueeze(1)

        # B: batch size
        # L: input frame length
        # L': padded input frame length
        # F: feature dimension
        # M: feature frame length
        # S: number of sources

        padded, num_pads = self._align_num_frames_with_strides(input)  # B, 1, L'
        batch_size, num_padded_frames = padded.shape[0], padded.shape[2]
        feats = self.encoder(padded)  # B, F, M
        masked = self.mask_generator(feats) * feats.unsqueeze(1)  # B, S, F, M
        masked = masked.view(batch_size * self.num_sources, self.enc_num_feats, -1)  # B*S, F, M
        decoded = self.decoder(masked)  # B*S, 1, L'
        output = decoded.view(batch_size, self.num_sources, num_padded_frames)  # B, S, L'
        if num_pads > 0:
            output = output[..., :-num_pads]  # B, S, L

        return output, []


def conv_tasnet_base(num_sources: int = 2) -> ConvTasNet:
    r"""Builds non-causal version of :class:`~torchaudio.models.ConvTasNet`.

    The parameter settings follow the ones with the highest Si-SNR metirc score in the paper,
    except the mask activation function is changed from "sigmoid" to "relu" for performance improvement.

    Args:
        num_sources (int, optional): Number of sources in the output.
            (Default: 2)
    Returns:
        ConvTasNet:
            ConvTasNet model.
    """
    return ConvTasNet(
        num_sources=num_sources,
        enc_kernel_size=16,
        enc_num_feats=512,
        msk_kernel_size=3,
        msk_num_feats=128,
        msk_num_hidden_feats=512,
        msk_num_layers=8,
        msk_num_stacks=3,
        msk_activate="relu",
    )



from dataclasses import dataclass
from pathlib import Path

import librosa
from simple_parsing import Serializable
from torch.utils.data import Dataset


@dataclass
class SyntheticDatasetArgs(Serializable):
    """DNS Challenge Interspeech 2020 synthetic dataset.

    Note:
        The dataset must have the same number of noisy and clean audio files.
        The noisy and clean audio files have the different filenames.

    Args:
        dataset_root_dir: The root directory of the dataset.
        audio_sample_len_in_sec: The length of the audio samples in seconds.
        offset: The index of the first audio file to use.
        limit: The maximum number of audio files to use.
        sr: The sample rate of the audio files.
        noisy_subdir: The name of the subdirectory containing the noisy audio files.
        clean_subdir: The name of the subdirectory containing the clean audio files.
    """

    dataset_root_dir: str = "datasets/VoiceBank-Demand"
    audio_sample_len_in_sec: int = 4
    offset: int = -1
    limit: int = -1
    sr: int = 16000
    noisy_subdir: str = "noisy"
    clean_subdir: str = "clean"


class SyntheticDataset(Dataset):
    def __init__(self, args: SyntheticDatasetArgs):
        super().__init__()
        self.args = args

        dataset_root_dir = Path(args.dataset_root_dir).expanduser().resolve()

        self.noisy_root_dir = dataset_root_dir / args.noisy_subdir
        self.clean_root_dir = dataset_root_dir / args.clean_subdir

        if not self.noisy_root_dir.exists():
            raise FileNotFoundError(f"Directory not found: {self.noisy_root_dir}")

        if not self.clean_root_dir.exists():
            raise FileNotFoundError(f"Directory not found: {self.clean_root_dir}")

        self.noisy_audio_fpath_list = librosa.util.find_files(self.noisy_root_dir)
        self.clean_audio_fpath_list = librosa.util.find_files(self.clean_root_dir)

        assert len(self.noisy_audio_fpath_list) == len(self.clean_audio_fpath_list), (
            f"Number of noisy and clean audio files must be equal. "
            f"Found {len(self.noisy_audio_fpath_list)} noisy files and "
            f"{len(self.clean_audio_fpath_list)} clean files."
        )

        if args.offset > 0:
            self.noisy_audio_fpath_list = self.noisy_audio_fpath_list[args.offset :]

        if args.limit > 0:
            step = max(len(self.noisy_audio_fpath_list) // args.limit, 1)
            self.noisy_audio_fpath_list = self.noisy_audio_fpath_list[::step]

        if args.audio_sample_len_in_sec > 0:
            self.audio_sample_len_in_samples = int(args.audio_sample_len_in_sec * args.sr)

    def __len__(self):
        return len(self.noisy_audio_fpath_list)

    def __getitem__(self, idx):
        noisy_fpath = self.noisy_audio_fpath_list[idx]
        stem = Path(noisy_fpath).stem
        file_id = stem.split("_")[-1]

        clean_fname = f"clean_fileid_{file_id}.wav"
        clean_fpath = (self.clean_root_dir / clean_fname).as_posix()

        noisy_y, _ = librosa.load(noisy_fpath, sr=self.args.sr)
        clean_y, _ = librosa.load(clean_fpath, sr=self.args.sr)

        return noisy_y, clean_y, stem



from itertools import permutations

import torch


class PairwiseNegSDR:
    def __init__(self, zero_mean=True, EPS=1e-8):
        self.zero_mean = zero_mean
        self.EPS = EPS

    def __call__(self, est, ref):
        """_summary_

        Args:
            est (`torch.Tensor` of shape [batch_size, num_sources, sequence_length]):
                The batch of target estimates.
            ref: _description_

        Returns:
            `torch.Tensor` of shape [batch_size, num_sources, num_sources]:
                The pairwise losses between all permutations of the sources. Dim 1 corresp. to estimates and dim 2 to references.
        """
        if ref.shape != est.shape or ref.ndim != 3:
            raise TypeError(f"Inputs must be of shape [batch, n_src, time], got {ref.shape} and {est.shape} instead")

        # Zero-mean norm for removing DC offset
        if self.zero_mean:
            mean_ref = torch.mean(ref, dim=2, keepdim=True)
            mean_est = torch.mean(est, dim=2, keepdim=True)
            ref = ref - mean_ref
            est = est - mean_est

        # Pair-wise SI-SDR. (Reshape to use broadcast)
        s_est = torch.unsqueeze(est, dim=2)  # [batch, n_src, 1, time]
        s_ref = torch.unsqueeze(ref, dim=1)  # [batch, 1, n_src, time]

        # [batch, n_src, n_src, 1]
        pair_wise_dot = torch.sum(s_est * s_ref, dim=3, keepdim=True)

        # [batch, 1, n_src, 1]
        s_ref_energy = torch.sum(s_ref**2, dim=3, keepdim=True) + self.EPS

        # project est on target
        # [batch, n_src, n_src, time]
        pair_wise_proj = pair_wise_dot * s_ref / s_ref_energy

        # Distortion
        e_noise = s_est - pair_wise_proj

        # SDR
        # [batch, n_src, n_src]
        pair_wise_sdr = torch.sum(pair_wise_proj**2, dim=3) / (torch.sum(e_noise**2, dim=3) + self.EPS)

        pair_wise_sdr = 10 * torch.log10(pair_wise_sdr + self.EPS)

        return -pair_wise_sdr


class PITWrapper:
    def __init__(self, loss_func):
        self.loss_func = loss_func

    def find_best_perm(self, pair_wise_losses):
        """Find the best permutation using the simplest method.

        Args:
            pw_losses (`torch.Tensor` of shape [batch_size, num_sources, num_sources]):
                The pairwise losses between all permutations of the sources.
        """
        num_sources = pair_wise_losses.shape[1]

        # After transposition, dim 1 corresp. to sources and dim 2 to estimates
        pwl = pair_wise_losses.transpose(-1, -2)

        # Create all possible permutations of the sources
        # e.g., [(0, 1), (1, 0)] with num_sources = 2 (shape = [2!, 2, 1])
        perms = pwl.new_tensor(list(permutations(range(num_sources))), dtype=torch.long)

        # Column permutation indices
        idx = perms.unsqueeze(2)

        # Loss mean of each permutation
        # one-hot, [n_src!, n_src, n_src]
        perms_one_hot = pwl.new_zeros((*perms.size(), num_sources)).scatter_(2, idx, 1)
        loss_set = torch.einsum("bij,pij->bp", [pwl, perms_one_hot])
        loss_set /= num_sources

        # Indexes and values of min losses for each batch element
        min_loss, min_loss_idx = torch.min(loss_set, dim=1)

        # Permutation indices for each batch.
        batch_indices = torch.stack([perms[m] for m in min_loss_idx], dim=0)

        return min_loss, batch_indices

    @staticmethod
    def reorder_source(source, batch_indices):
        r"""Reorder sources according to the best permutation.

        Args:
            source (torch.Tensor): Tensor of shape :math:`(batch, n_src, time)`
            batch_indices (torch.Tensor): Tensor of shape :math:`(batch, n_src)`.
                Contains optimal permutation indices for each batch.
        """
        reordered_sources = torch.stack([torch.index_select(s, 0, b) for s, b in zip(source, batch_indices)])
        return reordered_sources

    def __call__(self, est, ref, **kwargs):
        """

        Args:
            est (`torch.Tensor` of shape [batch_size, num_sources, ...]):
                The batch of target estimates.
            ref: _description_
        """
        num_sources = est.shape[1]
        assert num_sources < 10, f"Expected source axis along dim 1, found {num_sources}"

        # [batch_size, num_sources, num_sources]
        pw_losses = self.loss_func(est, ref, **kwargs)
        min_loss, batch_indices = self.find_best_perm(pw_losses)
        mean_loss = torch.mean(min_loss)
        reordered = self.reorder_source(est, batch_indices)
        return mean_loss, reordered


if __name__ == "__main__":
    ref = torch.rand(2, 3, 16000)
    est = torch.ones(2, 3, 16000)

    loss_func = PairwiseNegSDR(zero_mean=False)
    pit_loss = PITWrapper(loss_func)
    loss, reordered = pit_loss(est, ref)
    print(loss)



from typing import Tuple

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn import Parameter


class SubBandLSTMCell(nn.Module):
    def __init__(self, input_size, hidden_size, sub_band_size, alpha):
        super(SubBandLSTMCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.alpha = alpha
        self.sub_band_size = sub_band_size

        self.weight_x = Parameter(torch.randn(4 * hidden_size, sub_band_size))
        self.weight_h = Parameter(torch.randn(4 * hidden_size, alpha * sub_band_size))
        self.bias_ih = Parameter(torch.randn(4 * hidden_size))
        self.bias_hh = Parameter(torch.randn(4 * hidden_size))

    def forward(self, input: Tensor, state: Tuple[Tensor, Tensor]) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        """
        Args:
            input: [B, N, F]
            state: ([B, aN], [B, H])

        Notes:
            input: [B, N, F]
            hidden_state: (h_x: [B, aN], c_x: [B, aN]), 每个样本都会有一个自己的 hidden state
            x_t: [B, F, N]
            W_x: [H, N]
            W_h: [H, H]


        Returns:
            [B, H], ([B, H], [B, H])
        """
        hx, cx = state
        batch_size, num_freqs, sub_band_size = input.shape

        # W_x @ x_t
        # [B, F, N] @ [N, H] => [B, F, H]
        # [N, H] 对于所有频带均共享
        # [B, F, N] => [BF, N]
        input = input.reshape(batch_size * num_freqs, sub_band_size)
        # [BF, N] @ [N, H] = [BF, H], 这里的 H 包含 4 个部分，但是每个部分是对于全频带共享权重的，维度为 [4H, N]
        torch.mm(input, self.weight_x.t())

        # W_h @ h_{t-1}
        # [B, aN] => [B, F, aN]
        # [B, F, aN] @ [aN, H] = [B, H]
        hx = hx[:, None, :].expand(-1, num_freqs, -1).reshape(batch_size * num_freqs, self.alpha * self.sub_band_size)

        gates = torch.mm(input, self.weight_x.t()) + self.bias_ih + torch.mm(hx, self.weight_h.t()) + self.bias_hh
        ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        cx = cx[:, None, :].expand(-1, num_freqs, -1).reshape(batch_size * num_freqs, -1)

        # 这里的 cx 限制了，难道 cx 单独修改一下？
        cy = (forgetgate * cx) + (ingate * cellgate)
        hy = outgate * torch.tanh(cy)

        hy = hy.reshape(batch_size, num_freqs, self.alpha * self.sub_band_size)
        cy = cy.reshape(batch_size, num_freqs, -1)

        return hy, (hy, cy)


class SubBandGRUCell(nn.Module):
    def __init__(self, input_size: int, sub_band_size: int, hidden_size: int, alpha: int):
        """

        Args:
            input_size: The number of expected features in the input `x`
            hidden_size: The number of features in the hidden state `h`
            sub_band_size: The number of features in each frequency
            alpha: The value of

        Notations:
            - H: hidden_state
            - ...

        Inputs: input, hidden_state
            - input of shape [B, F, N]: tensor containing input features
            - hidden_state of shape [B, F, aN]: ...

        Output: new_hidden_state
            - new_hidden_state of shape [B, F, aN]

        Notes:
            - [H, aN] 对于每个样本，每个频带都是共享的，但是每个频带都有它自己的输出（隐状态），这个可不是共享的
        """
        super().__init__()
        # for reset_gate and update gate
        self.weight_x = Parameter(torch.randn(2 * hidden_size, input_size))  # first term: W_x @ x_t
        self.weight_h = Parameter(torch.randn(2 * hidden_size, alpha * sub_band_size))  # second term: W_h @ h_{t-1}
        self.bias_x = Parameter(torch.randn(2 * hidden_size))
        self.bias_h = Parameter(torch.randn(2 * hidden_size))

        # for new gate
        self.weight_x_new_gate = Parameter(torch.randn(hidden_size, input_size))
        self.weight_h_new_gate = Parameter(torch.randn(hidden_size, alpha * sub_band_size))
        self.bias_x_new_gate = Parameter(torch.randn(hidden_size))
        self.bias_h_new_gate = Parameter(torch.randn(hidden_size))

        # from [B, H] to [B, aN]
        # e.g. [B, 256] to [B, 3 * 30]
        self.linear_update_gate = nn.Linear(hidden_size, alpha * sub_band_size, bias=False)
        self.linear_new_gate = nn.Linear(hidden_size, alpha * sub_band_size, bias=False)

        # export useful vars
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.alpha = alpha
        self.sub_band_size = sub_band_size

    def forward(self, input: Tensor, hidden_state: Tensor) -> Tensor:
        batch_size, num_freqs, sub_band_size = input.shape

        input = input.reshape(batch_size * num_freqs, sub_band_size)
        hidden_state = hidden_state.reshape(batch_size * num_freqs, self.alpha * self.sub_band_size)
        print(input.shape, hidden_state.shape)

        # reset_gate and update_gate
        # first "mm" term (W_x @ x_t): [BF, H] @ [N, H] => [BF, H]. [N, H] 对于所有频带均共享
        # second "mm" term (W_h @ h_{t-1}): [BF, aN] @ [aN, H] => [BF, H]
        # H 为存储了特征的变换信息，aN 存储了特征的记忆信息。
        # 特征的记忆信息可能并不需要那么多。我们将其从 H 变为 aN，与子带的大小相关，进而降低计算量。
        gates = (
            torch.mm(input, self.weight_x.t()) + self.bias_x + torch.mm(hidden_state, self.weight_h.t()) + self.bias_h
        )
        reset_gate, update_gate = gates.chunk(2, 1)
        reset_gate = torch.sigmoid(reset_gate)
        update_gate = torch.sigmoid(update_gate)

        # new gate
        new_gate = torch.tanh(
            torch.mm(input, self.weight_x_new_gate.t())
            + self.bias_x_new_gate
            + reset_gate * torch.mm(hidden_state, self.weight_h_new_gate.t())
            + self.bias_h_new_gate
        )

        # new hidden state
        # linear_layer used to rescale H to aN
        new_hidden_state = (1 - self.linear_update_gate(update_gate)) * self.linear_new_gate(
            new_gate
        ) + self.linear_update_gate(update_gate) * hidden_state

        new_hidden_state = new_hidden_state.reshape(batch_size, num_freqs, self.alpha * self.sub_band_size)

        # [B, aN]
        return new_hidden_state


class SubBandGRULayer(nn.Module):
    def __init__(self, input_size: int, sub_band_size: int, hidden_size: int, alpha: int):
        """

        Args:
            input_size:
            sub_band_size:
            hidden_size:
            alpha:

        Inputs: input, hidden_state
            - input of shape [B, F, N, T]
            - input of shape [B, aN]
        """
        super().__init__()
        # export useful vars
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.alpha = alpha
        self.sub_band_size = sub_band_size

        self.cell = SubBandGRUCell(input_size, sub_band_size, hidden_size, alpha)

    def forward(self, input, hidden_state=None):
        batch_size, feature_size, sub_band_size, seq_len = input.shape
        if hidden_state is None:
            hidden_state = torch.zeros(batch_size, self.hidden_size, dtype=input.dtype, device=input.device)

        outputs = []
        for i in range(seq_len):
            hidden_state = self.cell(input[..., i], hidden_state)
            outputs.append(hidden_state)

        outputs = torch.stack(outputs, dim=-1)

        return outputs, hidden_state


class SubBandGRU(nn.Module):
    def __init__(
        self,
        input_size: int,
        sub_band_size: int,
        hidden_size: int,
        alpha: int,
        num_layers: int,
    ):
        """

        Args:
            input_size:
            sub_band_size:
            hidden_size:
            alpha:
            num_layers:

        Inputs: input, hidden_state
            - input of shape [B, F, N, T]
            - input of shape [B, aN]

        Outputs:
            hidden_state: [num_layers, B, aN]
        """
        super().__init__()
        # export useful vars
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.alpha = alpha
        self.sub_band_size = sub_band_size
        self.num_layers = num_layers

        self.layers = [
            SubBandGRULayer(input_size, sub_band_size, hidden_size, alpha),
        ]
        self.layers += [
            SubBandGRULayer(sub_band_size * alpha, sub_band_size, sub_band_size * alpha, alpha)
            for _ in range(num_layers - 1)
        ]
        self.layers = nn.ModuleList(self.layers)

    def forward(self, input, hidden_state=None):
        batch_size, feature_size, sub_band_size, seq_len = input.shape

        if hidden_state is None:
            hidden_state = torch.zeros(
                self.num_layers,
                batch_size,
                feature_size,
                self.alpha * self.sub_band_size,
                dtype=input.dtype,
                device=input.device,
            )

        i = 0
        output = input
        output_states = []
        for layer in self.layers:
            print("input", i, output.shape, hidden_state[i].shape)
            output, output_state = layer(output, hidden_state[i])
            print("output", i, output.shape, hidden_state[i].shape)
            output_states += [output_state]
            i += 1

        output_states = torch.stack(output_states, dim=0)
        return output, output_states


if __name__ == "__main__":
    batch_size = 2
    num_freqs = 257
    hidden_size = 256
    sub_band_size = 3
    alpha = 10
    num_frames = 10
    num_layers = 3

    ipt = torch.rand(batch_size, num_freqs, sub_band_size, num_frames)
    hx = torch.rand(num_layers, batch_size, num_freqs, sub_band_size * alpha)
    model = SubBandGRU(
        input_size=sub_band_size,
        sub_band_size=sub_band_size,
        hidden_size=hidden_size,
        alpha=alpha,
        num_layers=num_layers,
    )
    print(model(ipt)[0].shape)



import json
import os


os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
import subprocess
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from huggingface_hub import ModelCard, create_repo, upload_file, upload_large_folder, whoami

from audiozen.trainer_utils import collect_env


@dataclass
class Args:
    # The path to the model folder
    model_folder_path: str = "/path/to/your/model"
    # Whether the repo should be private
    private: bool = True
    # Whether to only upload `exp` folder
    only_exp_folder: bool = False
    # Whether to only upload the best checkpoint. If True, all other checkpoints will be ignored.
    only_best_checkpoint: bool = True
    # Whether to create a model card, or upload the README.md file if it exists
    create_model_card: bool = True


def generate_repo_id(model_folder_path: Path) -> str:
    """recipes/dns_icassp_2023/**/bsrnn => dns_icassp_2023_**_bsrnn."""
    # Get username
    user = whoami()["name"]

    path_parts = model_folder_path.parts
    # Get all parts after "recipes"
    valid_path_parts = path_parts[path_parts.index("recipes") + 1 :]

    repo_name = "__".join(valid_path_parts)
    repo_id = f"{user}/{repo_name}"

    print(f"Will upload model path {model_folder_path.as_posix()} to repo id {repo_id}")
    return repo_id


def get_git_revision_hash(model_folder_path: Path) -> str:
    return subprocess.check_output(["git", "-C", model_folder_path, "rev-parse", "HEAD"]).strip().decode("utf-8")


def push_to_huggingface_hub(args: Args):
    """Push a model to the Hugging Face Hub.

    Args:
        args (Args): The arguments to push the model to the Hugging Face Hub.

    Notes:
        The model will be pushed to the Hugging Face Hub with the following structure:
        ```
        recipes/
            path/
                to/
                    model/
                        README.md
                        exp/
                        other_files/
        ```
    """
    model_folder_path = Path(args.model_folder_path).absolute()
    assert model_folder_path.exists(), f"Model folder {model_folder_path} does not exist."

    # Create a new repo on the hugingface hub if it doesn't exist
    repo_id = generate_repo_id(model_folder_path)
    repo_url = create_repo(repo_id=repo_id, token=None, private=args.private, repo_type="model", exist_ok=True)
    print(f"Created repo at `{repo_url}`")

    # Create a model card (README.md)
    # First check if the user wants to create a model card
    # If yes, create a model card with the following content:
    # If no, check if a README.md file exists in the model folder, and upload it to the repo
    if args.create_model_card:
        print(f"> Creating a auto-generated README.md for {repo_id}")

        model_card_content = f"""
---
language: en
license: mit
---
# {repo_id.split("/")[-1]}

This project was initially created for personal use and is now archived.
While the code and documentation may be incomplete, I'm happy to share it in case others find it useful.

This model was uploaded from AudioZen at **{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}**.

## Overview

- **Training folder:** `{model_folder_path.as_posix()}`
- **Git hash:** `{get_git_revision_hash(model_folder_path)}`

## Usage

Refer to https://huggingface.co/docs/hub/en/models-downloading for instructions on how to download this model.
For example, you can use the following code to download the model:

```bash
# Install git lfs if you haven't already
git lfs install

# Go into the directory where you want to download the model
cd {model_folder_path.parent.as_posix()}

# Clone the model checkpoint
git clone https://huggingface.co/{repo_id}
# or
git clone git@hf.co:/{repo_id}
```

## Model Architecture

- Model architecture:
- Hyperparameters: Check the `config.yaml` file in the training folder for hyperparameters.


## Environment Info

<details>

{collect_env()}

Note: This information was generated automatically when the model was uploaded instead of when the model was trained.
</details>

## Upload Arguments

<details>

```python
{json.dumps(vars(args), indent=4, sort_keys=True)}
```

</details>
        """

        model_card = ModelCard(model_card_content)
        model_card.save(model_folder_path / "modelcard.md")
        model_card.push_to_hub(repo_id=repo_id)
    else:
        print("> Not creating a ModelCard, uploading README.md if it exists...")
        if (model_folder_path / "README.md").exists():
            upload_file(path_or_fileobj=(model_folder_path / "README.md"), path_in_repo="README.md", repo_id=repo_id)
        else:
            print(f"> No README.md found in {model_folder_path}")

    # Upload model files and tensorboard logs
    print(f"> Uploading model files to `{repo_url}`...")

    upload_folder_kwargs = {
        "repo_id": repo_id,
        "repo_type": "model",
        "ignore_patterns": ["*__pycache__*", "*.nfs*"],
    }

    # If the user only wants to upload the `exp` folder, set the folder path to the `exp` folder
    if args.only_exp_folder:
        print("> Only uploading `exp` folder")
        upload_folder_kwargs["folder_path"] = (model_folder_path / "exp").as_posix()
        upload_folder_kwargs["path_in_repo"] = "exp"
    else:
        upload_folder_kwargs["folder_path"] = model_folder_path.as_posix()

    # If the user only wants to upload the best checkpoint, ignore all other checkpoints
    if args.only_best_checkpoint:
        print("> Only uploading the best checkpoint")
        upload_folder_kwargs["ignore_patterns"].append("**/epoch_*")
    else:
        # TODO Can We only upload the latest 10 checkpoints?
        print("> Uploading all checkpoints (best and `epoch_*` checkpoints)")

    upload_large_folder(**upload_folder_kwargs)

    print(f"Model uploaded to {repo_url}")



"""This module contains functions about Sound source localization and tracking (SSLT)"""

import torch


def cart2sph(cart, include_r=False):
    """Convert tensor-like cartesian coordinates to spherical coordinates.

    - cartesian: [x, y, z] => spherical: [elevation, azimuth, radius]
    - cartesian: [x, y] => spherical: [azimuth, radius]

    Args:
        cart: cartesian_coordinates
        include_r: the radius is optional

    Shapes:
        cart: [B, cartesian_coordinates, T], where cartesian_coordinates can be 2 (x,y) or 3 (x, y, z).
        sph: [B, spherical_coordinates, T], where spherical_coordinates can be 1 or 2.

    Returns:
        dim of spherical_coordinates can be 1 and 2.

    Note:
        Corresponding order relationship of cartesian and spherical coordinates are as follows:
        - 3D array: [x, y , z] <=> [elevation, azimuth, radius] <=> [theta, phi, r]
        - 2D array: [x, y] <=> [azimuth, radius] <=> [phi, r]
    """
    _, num_coordinates, _ = cart.shape  # [B, 2 or 3, T]
    assert num_coordinates == 2 or num_coordinates == 3, "Only support 2D or 3D coordinates."

    radius = torch.sqrt(torch.sum(torch.pow(cart, 2), dim=1))  # [B, 1, T]
    phi = torch.atan2(cart[:, 1, :], cart[:, 0, :])  # [B, 1, T]

    if num_coordinates == 2:
        sph = torch.stack((phi, radius), dim=-1) if include_r else phi
    else:
        theta = torch.acos(cart[:, 2, :] / radius)
        sph = torch.stack((theta, phi, radius), dim=-1) if include_r else torch.stack((theta, phi), dim=-1)

    return sph


def sph2cart(sph):
    """Tensor-like spherical coordinates to cartesian coordinates.

        1. sphere: [elevation, azimuth] => cartesian: [x, y, z]
        2. sphere: azimuth => cartesian: [x, y]

     Shapes:
        sph: [B, spherical_coordinates, T], where spherical_coordinates can be 1 or 2.
        cart: [B, cartesian_coordinates, T], where cartesian_coordinates can be 2 (x, y) or 3 (x, y, z).

    Returns:
         dim of spherical_coordinates can be 1, 2 or 3.

    Notes:
        Corresponding order relationship of cartesian and spherical coordinates are as follows:

            - 3D array: (x, y, z) <=> (elevation, azimuth, radius) <=> (theta, phi, r)
            - 2D array: (x, y) <=> (azimuth, radius) <=> (phi, r)

        The output is the **unity cartesian**, i.e., the radius is supposed to be 1.
    """
    _, num_coordinates, _ = sph.shape  # [B, 1 or 2, T]
    assert num_coordinates == 1 or num_coordinates == 2, "Only support 1D or 2D coordinates now."

    if num_coordinates == 1:
        x = torch.cos(sph[:, 0, :])
        y = torch.sin(sph[:, 0, :])
        return torch.stack((x, y), dim=1)
    else:
        x = torch.sin(sph[:, 0, :]) * torch.cos(sph[:, 1, :])
        y = torch.sin(sph[:, 0, :]) * torch.sin(sph[:, 1, :])
        z = torch.cos(sph[:, 0, :])
        return torch.stack((x, y, z), dim=1)



import torch.nn as nn
import torch.nn.functional as F

from audiozen.constant import EPSILON


class CausalConformerConvBlock(nn.Module):
    def __init__(
        self,
        feat_dim,
        kernel_size=32,
        dropout=0.1,
    ):
        """Conformer Convolution Block.

        Args:
            feat_dim: the number of input features.
            kernel_size: the size of the convolving kernel. Defaults to 32 in the paper.
            dropout: the probability of an element to be zeroed. Defaults to 0.1 in the paper.

        Note:
            The input tensor should be in the shape of [B, F, T].
            We will change the layer norm to causal layer norm in the future.
        """
        super().__init__()
        assert kernel_size >= 3, "kernel_size must be larger than 3."

        self.kernel_size = kernel_size
        self.padding_size = kernel_size - 1

        self.pointwise_conv1 = nn.Conv1d(
            in_channels=feat_dim,
            out_channels=feat_dim * 2,  # for GLU
            kernel_size=1,
            stride=1,
            padding=0,
        )

        self.pointwise_conv2 = nn.Conv1d(
            in_channels=feat_dim,
            out_channels=feat_dim,
            kernel_size=1,
            stride=1,
            padding=0,
        )

        self.depthwise_conv = nn.Conv1d(
            in_channels=feat_dim,
            out_channels=feat_dim,
            kernel_size=kernel_size,
            stride=1,
            padding=self.padding_size,
            groups=feat_dim,
            bias=True,
        )

        self.norm = nn.LayerNorm(feat_dim, eps=EPSILON)

        self.activation = nn.SiLU()

        self.dropout = nn.Dropout(dropout)

    def forward(self, xs):
        """Forward.

        Args:
            x: [B, F, T]

        Returns:
            [B, F, T]
        """
        # Layernorm (sentence_length should be in the second dimension)
        input = xs.transpose(1, 2)  # [B, T, F]
        input = self.norm(input)  # [B, T, F]
        input = input.transpose(1, 2)  # [B, F, T]

        # Pointwise Conv
        input = self.pointwise_conv1(input)  # [B, 2 * F, T]

        # GLU
        input = F.glu(input, dim=1)  # [B, F, T]

        # 1D Depthwise Conv
        input = self.depthwise_conv(input)  # [B, F, T]
        input = input[:, :, : -self.padding_size]

        # Layernorm
        input = input.transpose(1, 2)  # [B, T, F]
        input = self.norm(input)  # [B, T, F]

        # Swish Activation
        input = self.activation(input)  # [B, T, F]

        # Pointwise Conv
        input = input.transpose(1, 2)  # [B, F, T]
        input = self.pointwise_conv2(input)  # [B, F, T]

        # Dropout
        input = self.dropout(input)  # [B, F, T]

        return input + xs


if __name__ == "__main__":
    import torch
    from torchinfo import summary

    x = torch.rand(2, 257, 100)
    block = CausalConformerConvBlock(feat_dim=257, kernel_size=3)
    # about 0.28 M and 42 M MACs
    summary(block, input_size=(2, 257, 100), device="cpu")
    y = block(x)
    print(y.shape)



import torch
from torch.nn import functional

from audiozen.acoustics.audio_feature import drop_band
from audiozen.models.base_model import BaseModel
from audiozen.models.module.sequence_model import SequenceModel


class Model(BaseModel):
    def __init__(
        self,
        num_freqs,
        look_ahead,
        sequence_model,
        fb_num_neighbors,
        sb_num_neighbors,
        fb_output_activate_function,
        sb_output_activate_function,
        fb_model_hidden_size,
        sb_model_hidden_size,
        norm_type="offline_laplace_norm",
        num_groups_in_drop_band=2,
        weight_init=True,
    ):
        """FullSubNet model (cIRM mask)

        Args:
            num_freqs: Frequency dim of the input
            look_ahead: Number of use of the future frames
            fb_num_neighbors: How much neighbor frequencies at each side from fullband model's output
            sb_num_neighbors: How much neighbor frequencies at each side from noisy spectrogram
            sequence_model: Chose one sequence model as the basic model e.g., GRU, LSTM
            fb_output_activate_function: fullband model's activation function
            sb_output_activate_function: subband model's activation function
            norm_type: type of normalization, see more details in "BaseModel" class
        """
        super().__init__()
        assert sequence_model in (
            "GRU",
            "LSTM",
        ), f"{self.__class__.__name__} only support GRU and LSTM."

        self.fb_model = SequenceModel(
            input_size=num_freqs,
            output_size=num_freqs,
            hidden_size=fb_model_hidden_size,
            num_layers=2,
            bidirectional=False,
            sequence_model=sequence_model,
            output_activate_function=fb_output_activate_function,
        )

        self.sb_model = SequenceModel(
            input_size=(sb_num_neighbors * 2 + 1) + (fb_num_neighbors * 2 + 1),
            output_size=2,
            hidden_size=sb_model_hidden_size,
            num_layers=2,
            bidirectional=False,
            sequence_model=sequence_model,
            output_activate_function=sb_output_activate_function,
        )

        self.sb_num_neighbors = sb_num_neighbors
        self.fb_num_neighbors = fb_num_neighbors
        self.look_ahead = look_ahead
        self.norm = self.norm_wrapper(norm_type)
        self.num_groups_in_drop_band = num_groups_in_drop_band

        if weight_init:
            self.apply(self.weight_init)

    def forward(self, noisy_mag):
        """
        Args:
            noisy_mag: noisy magnitude spectrogram

        Returns:
            The real part and imag part of the enhanced spectrogram

        Shapes:
            noisy_mag: [B, 1, F, T]
            return: [B, 2, F, T]
        """
        assert noisy_mag.dim() == 4
        noisy_mag = functional.pad(noisy_mag, [0, self.look_ahead])  # Pad the look ahead
        batch_size, num_channels, num_freqs, num_frames = noisy_mag.size()
        assert num_channels == 1, f"{self.__class__.__name__} takes the mag feature as inputs."

        # Fullband model
        fb_input = self.norm(noisy_mag).reshape(batch_size, num_channels * num_freqs, num_frames)
        fb_output = self.fb_model(fb_input).reshape(batch_size, 1, num_freqs, num_frames)

        # Unfold fullband model's output, [B, N=F, C, F_f, T]. N is the number of sub-band units

        fb_output_unfolded = self.freq_unfold(fb_output, num_neighbors=self.fb_num_neighbors)
        fb_output_unfolded = fb_output_unfolded.reshape(
            batch_size, num_freqs, self.fb_num_neighbors * 2 + 1, num_frames
        )

        # Unfold noisy spectrogram, [B, N=F, C, F_s, T]
        noisy_mag_unfolded = self.freq_unfold(noisy_mag, num_neighbors=self.sb_num_neighbors)
        noisy_mag_unfolded = noisy_mag_unfolded.reshape(
            batch_size, num_freqs, self.sb_num_neighbors * 2 + 1, num_frames
        )

        # Concatenation, [B, F, (F_s + F_f), T]
        sb_input = torch.cat([noisy_mag_unfolded, fb_output_unfolded], dim=2)
        sb_input = self.norm(sb_input)

        # Speeding up training without significant performance degradation.
        # These will be updated to the paper later.
        if batch_size > 1:
            sb_input = drop_band(
                sb_input.permute(0, 2, 1, 3), num_groups=self.num_groups_in_drop_band
            )  # [B, (F_s + F_f), F//num_groups, T]
            num_freqs = sb_input.shape[2]
            sb_input = sb_input.permute(0, 2, 1, 3)  # [B, F//num_groups, (F_s + F_f), T]

        sb_input = sb_input.reshape(
            batch_size * num_freqs,
            (self.sb_num_neighbors * 2 + 1) + (self.fb_num_neighbors * 2 + 1),
            num_frames,
        )

        # [B * F, (F_s + F_f), T] => [B * F, 2, T] => [B, F, 2, T]
        sb_mask = self.sb_model(sb_input)
        sb_mask = sb_mask.reshape(batch_size, num_freqs, 2, num_frames).permute(0, 2, 1, 3).contiguous()

        output = sb_mask[:, :, :, self.look_ahead :]
        return output


if __name__ == "__main__":
    import time

    from torchinfo import summary

    with torch.no_grad():
        noisy_mag = torch.rand(1, 1, 481, 100)
        model = Model(
            sb_num_neighbors=15,
            fb_num_neighbors=0,
            num_freqs=481,
            look_ahead=2,
            sequence_model="LSTM",
            fb_output_activate_function="ReLU",
            sb_output_activate_function=False,
            fb_model_hidden_size=512,
            sb_model_hidden_size=384,
            weight_init=False,
            norm_type="offline_laplace_norm",
            num_groups_in_drop_band=2,
        )
        start = time.time()
        output = model(noisy_mag)
        end = time.time()
        print(end - start)
        print(summary(model, (1, 1, 481, 100), device="cpu"))
